// graphics.cpp
// Revision 21-jan-2010

#ifdef __BORLANDC__
#pragma warn -8027
#endif

#include "graphics_impl.h"

#include "config_threads.h"


#include "sysvar.h"
#include "error.h"
#include "showerror.h"
#include "var.h"
#include "key.h"
#include "charset.h"
#include "util.h"
#include "trace.h"

#include <iostream>
#include <string>
#include <vector>
#include <algorithm>
#include <sstream>
#include <iomanip>
#include <memory>
#include <map>
#include <queue>

#include <string.h>
#include <limits.h>
#include <math.h>



#if defined __unix__ || defined __linux__
#include <unistd.h>
#endif

#ifdef HAVE_LOCALE_H
#include <locale.h>
#endif


using util::touch;
using util::to_string;
using util::dim_array;

using std::cerr;
using std::endl;

using std::auto_ptr;
using std::map;
using std::make_pair;
using std::swap;
using std::for_each;


namespace sysvar= blassic::sysvar;


namespace blassic {

namespace graphics_impl {


using blassic::graphics::ScreenNum;
using blassic::vartype::VarString;


std::string program_name;

std::string default_title ("blassic");

bool inited= false;
//bool window_created= false;

const int text_mode= 0, user_mode= -1;

bool graphics_mode_active= false;
int actualmode= text_mode;


#ifndef BLASSIC_HAS_GRAPHICS

void no_graphics_support ()
{
	if (showdebuginfo () )
		cerr << "This version of Blassic was compiled "
			"without graphics support" << endl;
}

#endif


#ifdef BLASSIC_HAS_GRAPHICS


class SystemMap {
public:
	static GraphicsSystem * get (const std::string & sysname);
	static void check_events ();
private:
	SystemMap ();
	~SystemMap ();
	typedef map <std::string, GraphicsSystem *> sysmap_t;
	sysmap_t sysmap;
	static SystemMap instance;

	class event_checker;
	friend class event_checker;
};

SystemMap SystemMap::instance;

SystemMap::SystemMap ()
{
}

SystemMap::~SystemMap ()
{
}

GraphicsSystem * SystemMap::get (const std::string & sysname)
{
	TRF;

	sysmap_t::iterator it= instance.sysmap.find (sysname);
	if (it != instance.sysmap.end () )
		return it->second;
	else
	{
		TRDEB ("Creating graphics system");
		if (sysname.empty () )
		{
			#ifdef BLASSIC_USE_WINDOWS_GRAPHICS

			GraphicsSystem * psystem= createSystemMsWin ();
			if (! psystem)
				throw ErrBlassicInternal;

			auto_ptr <GraphicsSystem> asystem (psystem);
			instance.sysmap.insert (make_pair (sysname, psystem) );
			asystem.release ();
			pcurrentsystem= psystem;
			return pcurrentsystem;

			#elif defined BLASSIC_USE_X

			std::string realname= check_display (sysname);
			it= instance.sysmap.find (realname);
			if (it != instance.sysmap.end () )
				return it->second;

			GraphicsSystem * psystem= createSystemXWin (realname);
			if (! psystem)
				throw ErrBlassicInternal;

			auto_ptr <GraphicsSystem> asystem (psystem);
			instance.sysmap.insert (make_pair (sysname, psystem) );
			asystem.release ();
			pcurrentsystem= psystem;
			return pcurrentsystem;

			#else

			throw ErrNoGraphics;

			#endif
		}
		else
		{
			#ifndef BLASSIC_USE_X

			throw ErrNotImplemented;

			#else

			pcurrentsystem= createSystemXWin (sysname);
			return pcurrentsystem;

			#endif
		}
	}
}

class SystemMap::event_checker {
public:
	void operator () (SystemMap::sysmap_t::value_type & v)
	{
		v.second->check_events ();
	}
};

void SystemMap::check_events ()
{
	for_each (instance.sysmap.begin (), instance.sysmap.end (),
		event_checker () );
}


GraphicsSystem * pcurrentsystem= 0;


GraphicsSystem * getcurrentsystem ()
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	if (! pcurrentsystem)
	{
		TRDEB ("Creating...");

		#if 0

		#ifdef BLASSIC_USE_WINDOWS_GRAPHICS
		createSystemMsWin ();
		#elif defined BLASSIC_USE_X
		createSystemXWin ();
		#else
		throw ErrNoGraphics;
		#endif

		#endif
		pcurrentsystem= SystemMap::get (std::string () );
	}

	return pcurrentsystem;

	#else

	throw ErrFunctionCall;

	#endif
}

void destroy_current_system ()
{
	delete pcurrentsystem;
}


class ScreenMap {
public:
	static Screen * get (ScreenNum scnum);
	static Screen * find (ScreenNum scnum);
	static Screen * check (ScreenNum scnum,
		const std::string & sysname);
	static Screen * create (ScreenNum scnum,const std::string & sysname);
	static ScreenNum find (Screen * pscreen);
private:
	ScreenMap ();
	~ScreenMap ();
	typedef map <ScreenNum, Screen *> scmap_t;
	scmap_t scmap;
	static ScreenMap instance;

	class value_is {
	public:
		value_is (Screen * pscreen_n) :
			pscreen (pscreen_n)
		{ }
		bool operator () (const scmap_t::value_type & v)
		{
			return v.second == pscreen;
		}
	private:
		Screen * pscreen;
	};
	friend class value_is;
};

ScreenMap ScreenMap::instance;

ScreenMap::ScreenMap ()
{
}

ScreenMap::~ScreenMap ()
{
}

Screen * ScreenMap::get (ScreenNum scnum)
{
	TRF;

	scmap_t::iterator it= instance.scmap.find (scnum);
	if (it != instance.scmap.end () )
		return it->second;
	else
	{
		TRDEB ("Creating screen");

		//GraphicsSystem & sys= * getcurrentsystem ();
		GraphicsSystem & sys= * SystemMap::get (std::string () );

		Screen * psc (sys.newScreen () );
		auto_ptr <Screen> apsc (psc);
		instance.scmap.insert (make_pair (scnum, psc) );
		apsc.release ();
		return  psc;
	}
}

Screen * ScreenMap::find (ScreenNum scnum)
{
	scmap_t::iterator it= instance.scmap.find (scnum);
	if (it != instance.scmap.end () )
		return it->second;
	else
		return 0;
}

Screen * ScreenMap::check (ScreenNum scnum,
	const std::string & sysname)
{
	TRF;

	scmap_t::iterator it= instance.scmap.find (scnum);
	if (it == instance.scmap.end () )
		return 0;
	else
	{
		Screen * pcurrent= it->second;
		if (& (pcurrent->get_system () ) == SystemMap::get (sysname) )
			return pcurrent;
		else
			return 0;
	}
}

Screen * ScreenMap::create (ScreenNum scnum,
	const std::string & sysname)
{
	scmap_t::iterator it= instance.scmap.find (scnum);
	if (it != instance.scmap.end () )
	{
		it->second->destroy ();
		instance.scmap.erase (it);
	}
	GraphicsSystem & sys= * SystemMap::get (sysname);
	Screen * pscreen= sys.newScreen ();
	auto_ptr <Screen> ascreen (pscreen);
	instance.scmap.insert (make_pair (scnum, pscreen) );
	ascreen.release ();
	return pscreen;
}

ScreenNum ScreenMap::find (Screen * pscreen)
{
	scmap_t & m= instance.scmap;
	scmap_t::iterator it= find_if (m.begin (), m.end (),
		value_is (pscreen) );
	if (it == m.end () )
		throw ErrBlassicInternal;
	return it->first;
}


Screen * pcurrentscreen= 0;


Screen * getcurrentscreen ()
{
	TRF;

	#ifndef BLASSIC_HAS_GRAPHICS

	no_graphics_support ();
	throw ErrFunctionCall;

	#else

	#if 0
	if (! pcurrentscreen)
	{
		TRMESSAGE (tr, "Creating...");

		pcurrentscreen= getcurrentsystem ()->newScreen ();
	}

	return pcurrentscreen;
	#endif

	if (! pcurrentscreen)
		pcurrentscreen= ScreenMap::get (0);
	return pcurrentscreen;

	#endif
}

void create_screen (int width, int height, RotateMode rotmode,
	Inkset inkset, int pen, int paper,
	ZoomValue charx, ZoomValue chary)
{
	TRF;

	getcurrentscreen ()->create (width, height, rotmode,
		inkset, pen, paper,
		charx, chary);
	//window_created= true;
}

void create_screen (ScreenNum scnum,
	int width, int height, RotateMode rotmode,
	Inkset inkset, int pen, int paper,
	ZoomValue charx, ZoomValue chary)
{
	TRF;

	ScreenMap::get (scnum)->create (width, height, rotmode,
		inkset, pen, paper,
		charx, chary);
	//window_created= true;
}

void destroy_screen ()
{
	TRF;

	getcurrentscreen ()->destroy ();
}


#endif
// BLASSIC_HAS_GRAPHICS


inline void requiregraphics ()
{
	#ifndef BLASSIC_HAS_GRAPHICS

	no_graphics_support ();
	throw ErrFunctionCall;

	#else

	if (! graphics_mode_active)
		throw ErrNoGraphics;

	#endif
}


#ifdef BLASSIC_USE_X

std::string getDISPLAY ()
{
	const char * strdisplay= getenv ("DISPLAY");
	if (strdisplay == NULL)
		return std::string ();
	else
		return strdisplay;
}

//std::string last_display;

#endif


void initialize_graphics ()
{
	TRF;

	// Does the real initialization of the graphics system.
	// It will not be called until a graphics mode ise
	// established.

	#ifdef BLASSIC_HAS_GRAPHICS

	ASSERT (! inited);

	#ifdef BLASSIC_USE_X

	//blassic::graphics_impl::createSystemXWin ();
	inited= true;

	#elif defined  BLASSIC_USE_WINDOWS_GRAPHICS

	//blassic::graphics_impl::createSystemMsWin ();
	inited= true;

	#endif

	#else
	// No BLASSIC_HAS_GRAPHICS

	#endif
}

void destroy_text_windows ();

void check_initialized ()
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	if (! inited)
		initialize_graphics ();
	else
	{
		#ifdef BLASSIC_USE_X
		// Testing, temporarily disabled.
		#if 0
		if (last_display != getDISPLAY () )
		{
			destroy_text_windows ();
			graphics::setmode (text_mode);
			graphics::uninitialize ();
			initialize_graphics ();
		}
		#endif
		#endif
	}

	#endif
}


#ifdef BLASSIC_HAS_GRAPHICS

struct SpecialMode {
	std::string name;
	Inkset inkset;
	int pen;
	int paper;
	int width;
	int height;
	TransformType transform;
	int zoomx;
	int zoomy;
	SpecialMode (const std::string & name, Inkset inkset,
			int pen, int paper, int width, int height,
			TransformType transform, int zoomx, int zoomy) :
		name (name), inkset (inkset),
		pen (pen), paper (paper), width (width), height (height),
		transform (transform), zoomx (zoomx), zoomy (zoomy)
	{ }
};

const SpecialMode specialmodes []= {
	SpecialMode ("spectrum", InkSpectrum,  0,  7, 256, 192,
		TrInvertY,  1, 1),
	SpecialMode ("cpc0",     InkCpc,       1,  0, 640, 400,
		TrInvertY,  4, 2),
	SpecialMode ("cpc1",     InkCpc,       1,  0, 640, 400,
		TrInvertY,  2, 2),
	SpecialMode ("cpc2",     InkCpc,       1,  0, 640, 400,
		TrInvertY,  1, 2),
	SpecialMode ("pcw",      InkStandard,  0, 15, 720, 248,
		TrIdentity, 1, 1),
	SpecialMode ("pcw2",     InkStandard,  0, 15, 720, 496,
		TrIdentity, 1, 2)
};


//void recreate_windows ();


void set_mode (blassic::graphics::ScreenNum scnum, int mode,
	int width, int height, TransformType transform,
	Inkset inkset, int pen, int paper,
	ZoomValue charx= 1, ZoomValue chary= 1)
{
	TRF;

	ASSERT (charx >= 1 && charx <= MAXZOOMTEXTX);
	ASSERT (chary >= 1 && chary <= MAXZOOMTEXTY);

	TRDEBS ("Width " << short (width) << ", height " << short (height) );

	if (mode != 0 && (width <= 0 || height <= 0) )
		throw ErrImproperArgument;

	if (mode != text_mode)
	{
		check_initialized ();
		if (! inited)
		{
			if (showdebuginfo () )
				cerr << "Graphics system not initialized" <<
					endl;
			throw ErrFunctionCall;
		}
	}

	sysvar::set16 (sysvar::GraphicsWidth, short (width) );
	sysvar::set16 (sysvar::GraphicsHeight, short (height) );

	RotateMode rotmode= RotateNone;
	switch (sysvar::get (sysvar::GraphRotate) )
	{
	case 0:
		// Nothing to do.
		break;
	case 1:
		rotmode= Rotate90;
		// Rotate 90 degrees.
		break;
	default:
		throw ErrFunctionCall;
	}

	TRDEBS ("Changing mode from " << actualmode << " to " << mode);

	if (mode != text_mode)
	{
		Screen & screen= * ScreenMap::get (scnum);

		//if (actualmode != text_mode)
		//	destroy_screen ();
		screen.destroy ();

		#if 0
		create_screen (scnum,
			width, height, rotmode,
			inkset, pen, paper,
			charx, chary);
		#endif
		screen.create (width, height, rotmode,
			inkset, pen, paper,
			charx, chary);

		graphics_mode_active= true;
		screen.set_transform (transform);
	}
	else
	{
		if (actualmode != text_mode)
		{
			Screen & screen= * ScreenMap::get (scnum);
			screen.destroy ();
		}
		graphics_mode_active= false;
	}

	actualmode= mode;
}


#endif


} // namespace graphics_impl


namespace graphics {

using namespace graphics_impl;

}


} // namespace blassic



void blassic::graphics::initialize (const char * progname)
{
	TRF;

	// Default symbol after and charset initialization:
	symbolafter (0);

	#ifdef BLASSIC_HAS_GRAPHICS

	program_name= progname;

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (progname);

	#endif
}

void blassic::graphics::uninitialize ()
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	if (! inited)
		return;

	//if (graphics_mode_active)
	//	setmode (0);

	destroy_current_system ();

	inited= false;

	#endif
	// BLASSIC_HAS_GRAPHICS
}

void blassic::graphics::origin (int x, int y)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS
	getcurrentscreen ()->set_origin (x, y);
	#else
	touch (x, y);
	#endif
}

void blassic::graphics::limits (int minx, int maxx, int miny, int maxy)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS
	getcurrentscreen ()->set_limits (minx, maxx, miny, maxy);
	#else
	touch (minx, maxx, miny, maxy);
	#endif
}

void blassic::graphics::ink (int inknum, int cpccolor)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->setcpcink
		(static_cast <InkNum> (inknum), cpccolor);

	#else

	touch (inknum, cpccolor);

	#endif
}

void blassic::graphics::ink (int inknum, int r, int g, int b)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	ColorRGB rgb (r, g, b);
	getcurrentscreen ()->setink (static_cast <InkNum> (inknum), rgb);

	#else

	touch (inknum, r, g, b);

	#endif
}

void blassic::graphics::clearink ()
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->init_colors ();

	#endif
}

void blassic::graphics::cls ()
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->cls ();

	#endif
	// BLASSIC_HAS_GRAPHICS
}

void blassic::graphics::setborder (int color)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->set_border (color);

	#else

	touch (color);
	throw ErrNoGraphics;

	#endif
}

void blassic::graphics::setborderwidth (int width)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->set_border_width (width);

	#else

	touch (width);
	throw ErrNoGraphics;

	#endif
}

void blassic::graphics::setscreen (ScreenNum scnum)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	pcurrentscreen= ScreenMap::get (scnum);
	pcurrentsystem= & pcurrentscreen->get_system ();

	#else

	touch (scnum);
	throw ErrFunctionCall;

	#endif
}

void blassic::graphics::setscreen (ScreenNum scnum,
	const std::string & sysname)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	Screen * pcurrent= ScreenMap::check (scnum, sysname);
	if (! pcurrent)
		/*pcurrent=*/ ScreenMap::create (scnum, sysname);
	//pcurrentscreen= pcurrent;

	#else

	touch (scnum, sysname);
	throw ErrFunctionCall;

	#endif
}

void blassic::graphics::setmode (ScreenNum scnum,
	int width, int height, TransformType transform,
	Inkset inkset, int pen, int paper,
	ZoomValue charx, ZoomValue chary)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	blassic::graphics_impl::set_mode (scnum,
		user_mode, width, height, transform,
		inkset, pen, paper, charx, chary);

	#else

	touch (scnum, width, height, transform);
	touch (inkset, pen, paper, charx, chary);
	throw ErrFunctionCall;

	#endif
}

void blassic::graphics::setmode (ScreenNum scnum, int width, int height,
	bool inverty, int zoomx, int zoomy)
{
	TRF;

	#ifndef BLASSIC_HAS_GRAPHICS

	touch (scnum, width, height, inverty, zoomx, zoomy);
	no_graphics_support ();

	#else

	ZoomValue zx= zoomx;
	ZoomValue zy= zoomy;

	if (zx < 1 || zx > MAXZOOMTEXTX || zy < 1 || zy > MAXZOOMTEXTY)
	{
		if (showdebuginfo () )
			cerr << "Invalid char zoom value" << endl;
		throw ErrImproperArgument;
	}
	TransformType transform= inverty ? TrInvertY : TrIdentity;

	blassic::graphics_impl::set_mode (scnum,
		user_mode, width, height, transform,
		InkStandard, 0, 15,
		zx, zy);

	#endif
}

void blassic::graphics::setmode (ScreenNum scnum, int mode)
{
	TRF;

	#ifndef BLASSIC_HAS_GRAPHICS

	no_graphics_support ();
	touch (scnum, mode);

	#else

	int width, height;
	switch (mode)
	{
	case 1:
		width= 320; height= 200; break;
	case 2:
		width= 640; height= 200; break;
	case 5:
		width= 320; height= 200; break;
	case 10:
		width= 640; height= 480; break;
	case 11:
		width= 800; height= 600; break;
	default:
		width= 0; height= 0;
	}
	if (mode != 0 && width == 0)
	{
		if (showdebuginfo () )
			cerr << "Invalid mode number " << mode << endl;
		throw ErrFunctionCall;
	}

	blassic::graphics_impl::set_mode (scnum,
		mode, width, height, TrIdentity, InkStandard, 0, 15);

	#endif
}

void blassic::graphics::setmode (ScreenNum scnum,
	const std::string & mode)
{
	TRFDEBS ("mode: " << mode);

	#ifndef BLASSIC_HAS_GRAPHICS

	no_graphics_support ();
	touch (scnum, mode);

	#else

	for (size_t i= 0; i < dim_array (specialmodes); ++i)
	{
		const SpecialMode & m= specialmodes [i];
		if (m.name == mode)
		{
			//default_pen= m.pen;
			//default_paper= m.paper;

			blassic::graphics_impl::set_mode (scnum,
				user_mode,
				m.width, m.height, m.transform,
				m.inkset, m.pen, m.paper,
				m.zoomx, m.zoomy);

			Screen * pscreen=
				getcurrentscreen ();
			//activetransform= m.transform;

			// Spectrum is an special case.
			if (mode == "spectrum")
			{
				pscreen->set_origin (0, 16);
				pscreen->set_limits (0, 256, 16, 192);
			}
			return;
		}
	}

	TRDEB ("Invalid mode");
	if (showdebuginfo () )
		cerr << '\'' << mode << "' is not a valid graphics mode" <<
			endl;
	throw ErrFunctionCall;

	#endif
}

bool blassic::graphics::ingraphicsmode ()
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return graphics_mode_active;

	#else

	return false;

	#endif
}

bool blassic::graphics::ingraphicsmode (ScreenNum scnum)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	Screen * pscreen= ScreenMap::find (scnum);
	return pscreen && pscreen->is_active ();

	#else

	return false;

	#endif
}

void blassic::graphics::setcolor (int color)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->set_graphics_pen (color);

	#else

	// No BLASSIC_HAS_GRAPHICS
	touch (color);

	#endif
}

int blassic::graphics::getcolor ()
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	return getcurrentscreen ()->get_graphics_pen ();

	#else

	throw ErrFunctionCall;

	#endif
}

void blassic::graphics::setbackground (int color)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->set_graphics_paper (color);

	#else

	touch (color);

	#endif
}

int blassic::graphics::getbackground ()
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	return getcurrentscreen ()->get_graphics_paper ();

	#else

	return 0;

	#endif
}

void blassic::graphics::settransparent (int transpmode)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	if (! inited)
		return;

	getcurrentscreen ()->set_transparent (transpmode & 1);

	#else

	touch (transpmode);

	#endif
}

void blassic::graphics::setdrawmode (int mode)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	if (! inited)
		return;

	getcurrentscreen ()->setdrawmode (mode);

	#else

	// No graphics.
	touch (mode);

	#endif
}


#ifdef BLASSIC_HAS_GRAPHICS


namespace blassic {


namespace graphics_impl {


#if 0
void do_line_unmasked (int x, int y)
{
	#if 0
	Screen * pscreen= getcurrentscreen ();

	int prevx= lastx, prevy= lasty;
	lastx= x; lasty= y;

	pscreen->transform_x (x); pscreen->transform_x (prevx);
	pscreen->transform_y (y); pscreen->transform_y (prevy);

	do_rotate (prevx, prevy);
	do_rotate (x, y);

	//setactivecolor (pforeground);
	getcurrentscreen ()->active_foreground ();

	#ifdef BLASSIC_USE_X

	#if 1

	getcurrentscreen ()->raw_draw_line (prevx, prevy, x, y);

	#else
	if (! fSynchro)
		XDrawLine (display, window, gc, prevx, prevy, x, y);
	XDrawLine (display, pixmap, gcp, prevx, prevy, x, y);
	XFlush (display);
	#endif

	#elif defined BLASSIC_USE_WINDOWS_GRAPHICS

	if (! fSynchro)
	{
		MoveToEx (hdc, prevx, prevy, 0);
		LineTo (hdc, x, y);
		LineTo (hdc, x + 1, y); // Last point
	}

	MoveToEx (hdcPixmap, prevx, prevy, 0);
	LineTo (hdcPixmap, x, y);
	LineTo (hdcPixmap, x + 1, y); // Last point

	#endif

	#else

	getcurrentscreen ()->do_line_unmasked (x, y);

	#endif
}
#endif

#if 0
const unsigned char maskvaluedefault= '\xFF';
unsigned char maskvalue= maskvaluedefault;
bool maskdrawfirst= true;
unsigned maskpos= 0;
unsigned char auxmask []= { 1, 2, 4, 8, 16, 32, 64, 128 };
#endif

#if 0
inline void impl_plot_mask (int x, int y)
{
	if (maskvalue == maskvaluedefault)
		do_plot (x, y);
	else
	{
		if (maskvalue & auxmask [maskpos] )
			do_plot (x, y);
		if (++maskpos == 8) maskpos= 0;
	}
}
#endif


#if 0
void do_line_mask (int x, int y)
{
	int prevx= lastx, prevy= lasty;
	lastx= x; lasty= y;

	Screen * pscreen= getcurrentscreen ();

	pscreen->transform_x (prevx); pscreen->transform_x (x);
	pscreen->transform_y (prevy); pscreen->transform_y (y);

	//setactivecolor (pforeground);
	pscreen->active_foreground ();

	int px= x - prevx;
	int py= y - prevy;
	int d1x= px < 0 ? -1 : px > 0 ? 1 : 0;
	int d1y= py < 0 ? -1 : py > 0 ? 1 : 0;
	int d2x= d1x;
	int d2y= 0;
	int m= abs (px);
	int n= abs (py);
	if (m <= n)
	{
		d2x= 0;
		d2y= d1y;
		m= abs (py);
		n= abs (px);
	}
	int s= m / 2;
	for (int i= 0; i <= m; ++i)
	{
		if (i != 0 || maskdrawfirst)
		{
			//if (maskvalue & auxmask [maskpos] )
			//	do_plot (prevx, prevy);
			//if (++maskpos == 8) maskpos= 0;
			impl_plot_mask (prevx, prevy);
		}
		s+= n;
		if (s >= m)
		{
			s-= m;
			prevx+= d1x;
			prevy+= d1y;
		}
		else
		{
			prevx+= d2x;
			prevy+= d2y;
		}
	}
}
#endif

#if 0
inline void do_line (int x, int y)
{
	// When limited we use line masked to simplify line unmasked.
	if (! limited && (maskvalue == maskvaluedefault && maskdrawfirst) )
		do_line_unmasked (x, y);
	else
		do_line_mask (x, y);
}
#endif


#if 0
void do_drawarc (int x, int y, double g)
{
	getcurrentscreen ()->do_drawarc (x, y, g);
}
#endif


#ifdef BLASSIC_USE_X

#if 0
inline void do_rotate_in_char (int & x, int & y)
{
	#if 0
	// Desactivado temporalmente

	switch (rotate)
	{
	case RotateNone:
		break;
	case Rotate90:
		#if 1
		int newx= y;
		y= charwidth - x - 1;
		x= newx;
		#else
		int newy= x;
		x= charwidth - y - 1;
		y= newy;
		#endif
		break;
	}
	#else

	touch (x, y);

	#endif
}
#endif


#if 0
inline bool load_char_image (int x, int y, unsigned char (& ch) [8] )
{
	TRACEFUNC (tr, "load_char_image");

	class ImageGuard {
	public:
		ImageGuard (XImage * img) :
			img (img)
		{ }
		~ImageGuard ()
		{
			XDestroyImage (img);
		}
	private:
		XImage * img;
	};

	int width= charwidth, height= charheight;
	do_rotate_rel (width, height);
	do_rotate (x, y);
	switch (rotate)
	{
	case RotateNone:
		break;
	case Rotate90:
		y-= height - 1;
		break;
	}
	TRMESSAGE (tr, "at " + to_string (x) + ", " + to_string (y) );

	XImage * img= XGetImage (display, pixmap, x, y,
		width, height, AllPlanes, XYPixmap);
	if (img == NULL)
		throw ErrNoGraphics; // Not a good idea, provisional.

	ImageGuard guard (img);
	unsigned long back= XGetPixel (img, 0, 0);
	bool fFore= false;
	unsigned long fore= 0;
	for (int i= 0, ipos= 0; i < 8; ++i, ipos+= zoomtexty)
		for (int j= 0, jpos= 0; j < 8; ++j, jpos+= zoomtextx)
		{
			int rjpos= jpos, ripos= ipos;
			do_rotate_in_char (rjpos, ripos);
			unsigned long c= XGetPixel (img, rjpos, ripos);
			bool bit;
			if (c == back)
			{
				bit= false;
			}
			else
			{
				if (! fFore)
				{
					fFore= true;
					fore= c;
				}
				if (c != fore)
					return false;
				bit= true;
			}
			#if 1
			// Quitado provisionalmente
			if (zoomtextx > 1 || zoomtexty > 1)
			{
				int zx= zoomtextx, zy= zoomtexty;
				//do_rotate_rel (zx, zy);
				const int iend= ipos + zy;
				const int jend= jpos + zx;
				for (int ii= ipos; ii < iend; ++ii)
					for (int jj= jpos; jj < jend; ++jj)
					{
						int rj= jj, ri= ii;
						do_rotate_in_char (rj, ri);
						if (XGetPixel (img, rj, ri)
								!= c)
							return false;
					}
			}
			#endif
			ch [i]<<= 1;
			ch [i]|= bit;
		}
	//XDestroyImage (img);

	return true;
}
#endif


#elif defined BLASSIC_USE_WINDOWS_GRAPHICS


#if 0
inline bool load_char_image (int x, int y, unsigned char (& ch) [8] )
{
	//COLORREF back= GetPixel (hdcPixmap, x, y);
	int rx= x, ry= y;
	do_rotate (rx, ry);
	COLORREF back= GetPixel (hdcPixmap, rx, ry);
	bool fFore= false;
	COLORREF fore= 0;
	for (int i= 0, ipos= y; i < 8; ++i, ipos+= zoomtexty)
		for (int j= 0, jpos= x; j < 8; ++j, jpos+= zoomtextx)
		{
			//COLORREF c= GetPixel (hdcPixmap, jpos, ipos);
			int rjpos= jpos, ripos= ipos;
			do_rotate (rjpos, ripos);
			COLORREF c= GetPixel (hdcPixmap, rjpos, ripos);
			bool bit;
			if (c == back)
				bit= false;
			else
			{
				if (! fFore)
				{
					fFore= true;
					fore= c;
				}
				if (c != fore)
					//return  std::string ();
					return false;
				bit= true;
			}
			if (zoomtextx > 1 || zoomtexty > 1)
			{
				const int iend= ipos + zoomtexty;
				const int jend= jpos + zoomtextx;
				for (int ii= ipos; ii < iend; ++ii)
					for (int jj= jpos; jj < jend; ++jj)
					{
						int rj= jj, ri= ii;
						do_rotate (rj, ri);
						if (GetPixel (hdcPixmap,
								rj, ri) != c)
							return false;
					}
			}
			ch [i]<<= 1;
			ch [i]|= bit;
		}
	return true;
}
#endif


#else


#if 0
inline bool load_char_image (int, int, unsigned char (&) [8] )
{
	return false;
}
#endif


#endif


std::string copychrat (int x, int y, BlChar from, BlChar to)
{
	TRF;

	#if 0

	{
		std::ostringstream oss;
		oss << "at " << x << ", " << y;
		TRMESSAGE (tr, oss.str () );
	}

	unsigned char ch [8]= {0};

	//if (x < 0 || x > screenwidth - 8 || y < 0 || y > screenheight - 8)
	if (x < 0 || x > screenwidth - charwidth ||
		y < 0 || y > screenheight - charheight)
		return std::string ();

	if (! load_char_image (x, y, ch) )
		return std::string ();

	#ifndef NDEBUG
	{
		std::ostringstream oss;
		oss << "Char: " << std::hex;
		for (int i= 0; i < 8; ++i)
		{
			oss << std::setw (2) << std::setfill ('0') <<
				static_cast <unsigned int> (ch [i]) << ", ";
		}
		TRMESSAGE (tr, oss.str () );
	}
	#endif

	char chinv [8];
	for (int i= 0; i < 8; ++i)
		chinv [i]= static_cast <unsigned char> (~ ch [i]);

	//for (int i= 0; i < 256; ++i)
	int iFrom= static_cast <int> (from);
	int iTo= static_cast <int> (to) + 1;
	for (int i= iFrom; i < iTo; ++i)
	{
		if (memcmp (charset::data [i], ch, 8) == 0 ||
			memcmp (charset::data [i], chinv, 8) == 0)
		{
			return std::string (1, static_cast <char> (i) );
		}
	}

	return std::string ();

	#else

	return getcurrentscreen ()->copychrat (x, y, from, to);

	#endif
}


#if 0
// Deshabilitado temporalmente

class ColorTester {
	#ifdef BLASSIC_USE_X
	XImage * img;
	#define TYPETESTER 2
	#endif
public:
	ColorTester ()
	{
		#ifdef BLASSIC_USE_X
		#if TYPETESTER == 1
		img= XGetImage (display, pixmap,
			0, 0, 1, 1,
			AllPlanes, XYPixmap);
		#else
		int width= screenwidth, height= screenheight;
		do_rotate_rel (width, height);
		img= XGetImage (display, pixmap,
			0, 0, width, height,
			AllPlanes, XYPixmap);
		#endif
		if (img == NULL)
			throw ErrNoGraphics;
		#endif
	}
	~ColorTester ()
	{
		#ifdef BLASSIC_USE_X
		XDestroyImage (img);
		#endif
	}
	ColorValue operator () (int x, int y)
	{
		do_rotate (x, y);
		#ifdef BLASSIC_USE_X
		#if TYPETESTER == 1
		XGetSubImage (display, pixmap, x, y, 1, 1,
			AllPlanes,XYPixmap,
			img, 0, 0);
		return XGetPixel (img, 0, 0);
		#else
		return XGetPixel (img, x, y);
		#endif
		#elif defined BLASSIC_USE_WINDOWS_GRAPHICS
		return GetPixel (hdcPixmap, x, y);
		#endif
	}
};

#endif


#if 0
ColorValue do_test_color (int x, int y)
{

	do_rotate (x, y);

	#ifdef BLASSIC_USE_X

	XImage * img= XGetImage (display, pixmap, x, y, 1, 1,
		AllPlanes, XYPixmap);
	if (img == NULL)
		throw ErrNoGraphics; // Not a good idea, provisional.

	unsigned long color= XGetPixel (img, 0, 0);
	XDestroyImage (img);
	return color;

	#else

	return GetPixel (hdcPixmap, x, y);

	#endif
}
#endif


#if 0
int do_test (int x, int y)
{
	#if 0
	// Deshabilitado temporalmente

	//int x= lastx;
	//int y= lasty;
	//transform_x (x);
	//transform_y (y);

	if (x < 0 || x >= screenwidth || y < 0 || y >= screenheight)
		return 0;

	do_rotate (x, y);

	#ifdef BLASSIC_USE_X

	XImage * img= XGetImage (display, pixmap, x, y, 1, 1,
		AllPlanes, XYPixmap);
	if (img == NULL)
		throw ErrNoGraphics; // Not a good idea, provisional.

	unsigned long color= XGetPixel (img, 0, 0);
	XDestroyImage (img);
	for (int i= 0; i < 16; ++i)
		if (mapcolor (i).pc->pixel == color)
			return i;
	for (definedcolor_t::iterator it= definedcolor.begin ();
		it != definedcolor.end (); ++it)
	{
		if (it->second.pc->pixel == color)
			return it->first;
	}
	return -1;

	#elif defined BLASSIC_USE_WINDOWS_GRAPHICS

	COLORREF color= GetPixel (hdcPixmap, x, y);
	for (int i= 0; i < 16; ++i)
	{
		//const ColorRGB & c= assignRGB [i];
		const ColorRGB & c= mapcolor (i).rgb;
		if (RGB (c.r, c.g, c.b) == color)
		//if (GetNearestColor (hdcPixmap, RGB (c.r, c.g, c.b) ) ==
		//	color)
			return i;
	}
	for (definedcolor_t::iterator it= definedcolor.begin ();
		it != definedcolor.end (); ++it)
	{
		const ColorRGB & c= it->second.rgb;
		if (RGB (c.r, c.g, c.b) == color)
			return it->first;
	}
	return -1;

	#else

	return -1;

	#endif

	#else

	touch (x, y);
	return -1;

	#endif
}
#endif

#if 0
int test ()
{
	#if 0
	// Deshabilitado temporalmente

	int x= lastx;
	int y= lasty;
	transform_x (x);
	transform_y (y);
	return do_test (x, y);

	#else
	return -1;
	#endif
}
#endif


} // namespace graphics_impl


namespace graphics {

using namespace graphics_impl;

}


} // namespace blassic


#endif
//BLASSIC_HAS_GRAPHICS


void blassic::graphics::line (int x, int y)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->do_line (x, y);

	#else

	touch (x, y);

	#endif
}

void blassic::graphics::liner (int x, int y)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->do_line_rel (x, y);

	#else

	touch (x, y);

	#endif
}

void blassic::graphics::drawarc (int x, int y, double g)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	//do_drawarc (x, y, g);
	getcurrentscreen ()->do_drawarc (x, y, g);

	#else

	touch (x, y, g);

	#endif
}

void blassic::graphics::rectangle (Point org, Point dest)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	#if 0

	int x1= org.x;
	int y1= org.y;
	int x2= dest.x;
	int y2= dest.y;
	lastx= x2; lasty= y2;

	transform_x (x1); transform_x (x2);
	transform_y (y1); transform_y (y2);

	do_rotate (x1, y1);
	do_rotate (x2, y2);

	if (x1 > x2) swap (x1, x2);
	if (y1 > y2) swap (y1, y2);

	#ifdef BLASSIC_USE_WINDOWS_GRAPHICS

	HBRUSH hold= (HBRUSH) SelectObject (hdcPixmap, GetStockObject (NULL_BRUSH) );
	Rectangle (hdcPixmap, x1, y1, x2 + 1, y2 + 1);
	SelectObject (hdc, hold);
	if (! fSynchro)
	{

		hold= (HBRUSH) SelectObject (hdc, GetStockObject (NULL_BRUSH) );
		Rectangle (hdc, x1, y1, x2 + 1, y2 + 1);
		SelectObject (hdc, hold);
	}

	#elif defined BLASSIC_USE_X

	XDrawRectangle (display, pixmap, gcp,
		x1, y1, x2 - x1, y2 - y1);
	if (! fSynchro)
		XDrawRectangle (display, window, gc,
			x1, y1, x2 - x1, y2 - y1);

	#endif

	#else

	getcurrentscreen ()->rectangle (org, dest);

	#endif

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (org, dest);

	#endif
}

void blassic::graphics::rectanglefilled (Point org, Point dest)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	#if 0

	int x1= org.x;
	int y1= org.y;
	int x2= dest.x;
	int y2= dest.y;
	lastx= x2; lasty= y2;

	transform_x (x1); transform_x (x2);
	transform_y (y1); transform_y (y2);

	if (x1 > x2) swap (x1, x2);
	if (y1 > y2) swap (y1, y2);
	do_fill_rectangle (x1, y1, x2, y2, true);

	#else

	getcurrentscreen ()->fill_rectangle (org, dest);

	#endif

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (org, dest);

	#endif
}

void blassic::graphics::zxplot (Point p)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	#if 0
	int x= p.x / 2 + 1;
	int y= 21 - p.y / 2;
	p.x*= 4;
	p.y*= 4;
	Point p2 (p);
	p2.x+= 3;
	p2.y+= 3;
	rectanglefilled (p, p2);
	gotoxy (BlChannel (0), x, y);
	#endif

	getcurrentscreen ()->zx_plot (p);
	gotoxy (BlChannel (0), p.x / 2 + 1, 21 - p.y / 2);

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (p);

	#endif
}

void blassic::graphics::zxunplot (Point p)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	#if 0
	//getcurrentscreen ()->active_background ();
	zxplot (p);
	//getcurrentscreen ()->active_foreground ();
	#endif

	getcurrentscreen ()->zx_unplot (p);
	gotoxy (BlChannel (0), p.x / 2 + 1, 21 - p.y / 2);

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (p);

	#endif
}


void blassic::graphics::move (int x, int y)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->do_move (x, y);

	#else

	touch (x, y);

	#endif
}

void blassic::graphics::mover (int x, int y)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->do_move_rel (x, y);

	#else

	touch (x, y);

	#endif
}

void blassic::graphics::plot (int x, int y)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->plot (x, y);

	#else

	touch (x, y);

	#endif
}

void blassic::graphics::plotr (int x, int y)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->plot_rel (x, y);

	#else

	touch (x, y);

	#endif
}

int blassic::graphics::test (int x, int y, bool relative)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	// Check for graphics mode is done when calling move.
	#if 0
	if (relative)
	{
		x+= lastx;
		y+= lasty;
	}
	move (x, y);
	#endif

	#if 0
	if (relative)
		mover (x, y);
	else
		move (x, y);

	return blassic::graphics_impl::test ();
	#endif

	return getcurrentscreen ()->test (x, y, relative);

	#else

	no_graphics_support ();
	touch (x, y, relative);
	throw ErrBlassicInternal; // Make the compiler happy.

	#endif
}


#if 0
namespace {

void line_to_point (const blassic::graphics::Point & p)
{
	blassic::graphics::line (p.x, p.y);
}


} // namespace
#endif


void blassic::graphics::plot (std::vector <Point> & points)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	ASSERT (points.size () != 0);

	#if 0
	lastx= points [0].x;
	lasty= points [0].y;
	#endif

	//move (points [0].x, points [0].y);
	//std::for_each (points.begin () + 1, points.end (), line_to_point);

	Screen & sc= * getcurrentscreen ();
	const size_t l= points.size ();
	ASSERT (l > 0);
	sc.do_move (points [0].x, points [0].y);
	for (size_t i= 1; i < l; ++i)
		sc.do_line (points [i].x, points [i].y);

	#else

	// Must not come here, requiregraphics must throw.
	ASSERT (false);
	touch (points);

	#endif
}

void blassic::graphics::circle (int x, int y, int radius)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->circle (x, y, radius);

	#else
	// BLASSIC_HAS_GRAPHICS

	touch (x, y, radius);

	#endif
}

void blassic::graphics::arccircle (int x, int y, int radius,
	BlNumber arcbeg, BlNumber arcend)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->arccircle(x, y, radius, arcbeg, arcend);

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (x, y, radius, arcbeg, arcend);

	#endif
}

void blassic::graphics::ellipse (int ox, int oy, int rx, int ry)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->ellipse (ox, oy, rx, ry);

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (ox, oy, rx, ry);

	#endif
}

void blassic::graphics::arcellipse (int ox, int oy, int rx, int ry,
	BlNumber arcbeg, BlNumber arcend)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->arcellipse (ox, oy, rx, ry, arcbeg, arcend);

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (ox, oy, rx, ry, arcbeg, arcend);

	#endif
}


#ifdef BLASSIC_HAS_GRAPHICS


namespace {


#if 0
// Deshabilitado temporalmente


#if 0

class Painter {
	ColorValue paint, border;
	std::vector <std::vector <bool> > visited;
	ColorTester test;
public:
	Painter (ColorValue paint, ColorValue border) :
		paint (paint),
		border (border),
		visited (screenwidth, std::vector <bool> (screenheight) )
	{
	}

	bool check (int x, int y)
	{
		if (! check_limit (x, y) )
			return false;
		std::vector <bool>::reference v= visited [x] [y];
		if (v)
			return false;
		v= true;
		ColorValue c= test (x, y);
		if (c == border || c == paint)
			return false;
		return true;
	}

void do_paint_it (int x, int y/*, int xd, int yd*/)
{
	//if (! check_limit (x, y) )
	//	return;
	//int c= do_test (x, y);
	//if (c == borderattr || c == paintattr)
	//	return;

	do_plot (x, y);
	//visited [x] [y]= true;

	//ColorValue c;
	//if (xd != -1)
		#if 0
		for (int i=1; /*check_limit (x - i, y)*/ i < 2; ++i)
		{
			if (visited [x - i] [y] )
				break;
			visited [x - i] [y]= true;
			//c= do_test_color (x - i, y);
			c= test (x - i, y);
			if (c != border && c != paint)
				do_paint_it (x - i, y, 1, 0);
			else
				break;
		}
		#else
		if (check (x - 1, y) )
			do_paint_it (x - 1, y/*, 1, 0*/);
		#endif
	//if (xd != 1)
		#if 0
		for (int i= 1; /*check_limit (x + i, y)*/ i < 2; ++i)
		{
			if (visited [x + i] [y] )
				break;
			visited [x + i] [y]= true;
			//c= do_test_color (x + i, y);
			c= test (x + i, y);
			if (c != border && c != paint)
				do_paint_it (x + i, y, -1, 0);
			else
				break;
		}
		#else
		if (check (x + 1, y) )
			do_paint_it (x + 1, y/*, 1, 0*/);
		#endif
	//if (yd != -1)
		#if 0
		for (int i= 1; /*check_limit (x, y - i)*/ i < 2; ++i)
		{
			if (visited [x] [y - i] )
				break;
			visited [x] [y - i]= true;
			//c= do_test_color (x, y - i);
			c= test (x, y - i);
			if (c != border && c != paint)
				do_paint_it (x, y - i, 0, 1);
			else
				break;
		}
		#else
		if (check (x, y - 1) )
			do_paint_it (x, y - 1/*, 0, 1*/);
		#endif
	//if (yd != 1)
		#if 0
		for (int i= 1; /*check_limit (x, y + i)*/ i < 2; ++i)
		{
			if (visited [x] [y + i] )
				break;
			visited [x] [y + i]= true;
			//c= do_test_color (x, y + i);
			c= test (x, y + i);
			if (c != border && c != paint)
				do_paint_it (x, y + i, 0, -1);
			else
				break;
		}
		#else
		if (check (x, y + 1) )
			do_paint_it (x, y + 1/*, 0, -1*/);
		#endif
}

	void do_paint (int x, int y)
	{
		if (check (x, y) )
			do_paint_it (x, y/*, 0, 0*/);
	}

};

#elif 0

class Painter {
	ColorValue paint, border;
	std::vector <std::vector <bool> > visited;
	ColorTester test;
public:
	Painter (ColorValue paint, ColorValue border) :
		paint (paint),
		border (border),
		visited (screenwidth, std::vector <bool> (screenheight) )
	{
	}

//void paintN (int x, int y);
//void paintS (int x, int y);
//void paintE (int x, int y);
//void paintW (int x, int y);

void do_paint (int x, int y)
{
	if (! check_limit (x, y) )
		return;
	//ColorValue c= do_test_color (x, y);
	ColorValue c= test (x, y);
	if (c == paint || c == border)
		return;
	do_plot (x, y);
	visited [x] [y]= true;
	paintN (x, y - 1);
	paintS (x, y + 1);
	paintE (x + 1, y);
	paintW (x - 1, y);
}

void paintN (int x, int y)
{
	if (! check_limit (x, y) )
		return;
	if (visited [x] [y] )
		return;
	visited [x] [y]= true;
	//ColorValue c= do_test_color (x, y);
	ColorValue c= test (x, y);
	if (c == paint || c == border)
		return;
	do_plot (x, y);
	paintN (x, y - 1);
	paintE (x + 1, y);
	paintW (x - 1, y);
}

void paintS (int x, int y)
{
	if (! check_limit (x, y) )
		return;
	if (visited [x] [y] )
		return;
	visited [x] [y]= true;
	//ColorValue c= do_test_color (x, y);
	ColorValue c= test (x, y);
	if (c == paint || c == border)
		return;
	do_plot (x, y);
	paintS (x, y + 1);
	paintE (x + 1, y);
	paintW (x - 1, y);
}

void paintE (int x, int y)
{
	if (! check_limit (x, y) )
		return;
	if (visited [x] [y] )
		return;
	visited [x] [y]= true;
	//ColorValue c= do_test_color (x, y);
	ColorValue c= test (x, y);
	if (c == paint || c == border)
		return;
	do_plot (x, y);
	paintN (x, y - 1);
	paintS (x, y + 1);
	paintE (x + 1, y);
}

void paintW (int x, int y)
{
	if (! check_limit (x, y) )
		return;
	if (visited [x] [y] )
		return;
	visited [x] [y]= true;
	//ColorValue c= do_test_color (x, y);
	ColorValue c= test (x, y);
	if (c == paint || c == border)
		return;
	do_plot (x, y);
	paintN (x, y - 1);
	paintS (x, y + 1);
	paintW (x - 1, y);
}

};

#else

struct PPoint {
	int x;
	int y;
	PPoint () { }
	PPoint (int x, int y) : x (x), y (y) { }
	//PPoint (const PPoint & p) : x (p.x), y (p.y) { }
	bool check (ColorValue paint, ColorValue border,
		std::vector <std::vector <bool> > & visit,
		ColorTester & test)
	{
		if (! check_limit (x, y) )
			return false;
		std::vector <bool>::reference v= visit [x] [y];
		if (v)
			return false;
		v= true;
		//ColorValue c= do_test_color (x, y);
		ColorValue c= test (x, y);
		if (c == paint || c == border)
			return false;
		return true;
	}
	void plot () { do_plot (x, y); }
};

#if 0

class Painter {
	ColorValue paint, border;
	ColorTester test;
	struct seg {
		int y, xl, xr, dy;
		seg () { }
		seg (int y, int xl, int xr, int dy) :
			y (y), xl (xl), xr (xr), dy (dy)
		{ }
	};
	std::deque <seg> stack;
	bool check (int x, int y)
	{
		if (! check_limit (x, y) )
			return false;
		ColorValue c= test (x, y);
		if (c == paint || c == border)
			return false;
		return true;
	}
public:
	Painter (ColorValue paint, ColorValue border) :
		paint (paint),
		border (border)
	{
	}

void do_paint (int x, int y)
{
	seg s;
	int l, x1, x2, dy;
	if (! check_limit (x, y) )
		return;
	ColorValue ov;
	ov= test (x, y);
	if (ov == paint || ov == border)
		return;
	stack.push_back (seg (y,     x, x,  1) );
	stack.push_back (seg (y + 1, x, x, -1) );
	while (! stack.empty () )
	{
		s= stack.back ();
		stack.pop_back ();
		y= s.y + s.dy; x1= s.xl; x2= s.xr; dy= s.dy;
		for (x= x1; check (x, y); --x)
			do_plot (x, y);
		if (x >= x1)
			goto skip;
		l= x + 1;
		if (l < x1)
			stack.push_back (seg (y, l, x1 - 1, -dy) );
		x= x1 + 1;
		do {
			for ( ; check (x, y); ++x)
				do_plot (x, y);
			stack.push_back (seg (y, l, x - 1, dy) );
			if (x > x2 + 1)
				stack.push_back (seg (y, x2 + 1, x - 1, -dy) );
		skip:
			for (x++; x <= x2 && ! test (x, y); ++x)
				continue;
			l= x;
		} while (x <= x2);
	}
}

};


#elif 0

class Painter {
	ColorValue paint, border;
	ColorTester test;
public:
	Painter (ColorValue paint, ColorValue border) :
		paint (paint),
		border (border)
	{
	}

void do_paint (int x, int y)
{
	using std::vector;
	vector <vector <bool> >
		visit (screenwidth, vector <bool> (screenheight) );
	std::deque <PPoint> stack;
	PPoint pt, pn;
	stack.push_back (PPoint (x, y) );
	while (! stack.empty () )
	{
		pt= stack.back ();
		stack.pop_back ();

		pt.plot ();
		pn= PPoint (pt.x, pt.y + 1);
		if (pn.check (paint, border, visit, test) )
			stack.push_back (pn);

		pn= PPoint (pt.x, pt.y - 1);
		if (pn.check (paint, border, visit, test) )
			stack.push_back (pn);

		pn= PPoint (pt.x + 1, pt.y);
		if (pn.check (paint, border, visit, test) )
			stack.push_back (pn);

		pn= PPoint (pt.x - 1, pt.y);
		if (pn.check (paint, border, visit, test) )
			stack.push_back (pn);
	}
}

};

#else

class Painter {
	ColorValue paint, border;
	std::vector <std::vector <bool> > visited;
	std::deque <PPoint> processlist, nextlist;
	ColorTester test;
public:
	Painter (ColorValue paint, ColorValue border) :
		paint (paint),
		border (border),
		visited (screenwidth, std::vector <bool> (screenheight) )
	{
	}
	bool check (int x, int y)
	{
		if (! check_limit (x, y) )
			return false;
		std::vector <bool>::reference v= visited [x] [y];
		if (v)
			return false;
		v= true;
		ColorValue c= test (x, y);
		if (c == paint || c == border)
			return false;
		return true;
	}
	void gf (PPoint p)
	{
		//if (! p.check (paint, border, visited, test) )
		if (! check (p.x, p.y) )
			return;
		p.plot ();
		nextlist.push_back ( PPoint (p.x, p.y - 1) );
		nextlist.push_back ( PPoint (p.x, p.y + 1) );
		nextlist.push_back ( PPoint (p.x - 1, p.y) );
		nextlist.push_back ( PPoint (p.x + 1, p.y) );

	}
	void do_paint (int x, int y)
	{
		processlist.push_back (PPoint (x, y) );
		while (! processlist.empty () )
		{
			nextlist.clear ();
			while (! processlist.empty () )
			{
				PPoint cp= processlist.back ();
				processlist.pop_back ();
				gf (cp);
			}
			processlist= nextlist;
		}
	}
};

#endif

#endif


#endif


} // namespace

#endif
// BLASSIC_HAS_GRAPHICS


void blassic::graphics::paint (int x, int y, int paintattr, int borderattr)
{
	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	#if 0
	transform_x (x);
	transform_y (y);

	ColorValue paint= getColorValue (mapcolor (paintattr) );
	ColorValue border= getColorValue (mapcolor (borderattr) );

	setactivecolor (mapcolor (paintattr).pc);
	//do_paint (x, y /*, 0, 0*/, paint, border);
	Painter p (paint, border);
	p.do_paint (x, y /*, 0, 0*/ );
	setactivecolor (pforeground);
	#endif

	Screen & screen= * getcurrentscreen ();

	//Painter * ppainter= screen.getPainter (paintattr, borderattr);
	//ppainter->paint (x, y);
	//delete ppainter;
	screen.paint (paintattr, borderattr, x, y);
	screen.flush ();

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (x, y, paintattr, borderattr);

	#endif
}

void blassic::graphics::mask (int m)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->set_mask (static_cast <unsigned char> (m) );

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (m);

	#endif
}

void blassic::graphics::maskdrawfirstpoint (bool f)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	#if 0
	maskdrawfirst= f;
	#endif

	getcurrentscreen ()->set_maskfirstpoint (f);

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (f);

	#endif
}


#ifdef BLASSIC_HAS_GRAPHICS


namespace blassic {

namespace graphics_impl {


int symbol_after_is;


inline bool iscontrolchar (char c)
{
	//return iscntrl (static_cast <unsigned char> (c) );
	unsigned char c1= static_cast <unsigned char> (c);
	return c1 < 32;
}


class BlTextWindow {
public:
	BlTextWindow (Screen & sc);
	BlTextWindow (Screen & sc, int x1, int x2, int y1, int y2);

	Screen * get_screen ();

	void set (int x1, int x2, int y1, int y2);
	void setdefault ();
	void defaultcolors ();

	void synchronize_suspend ();
	void synchronize_restart ();

	void setinverse (bool active);
	void setbright (bool active);
	bool getinverse () const { return inversemode; }
	bool getbright () const { return bright; }

	int getwidth () const { return width; }
	int getxpos () const { return x; }
	int getypos () const { return y; }
	void gotoxy (int nx, int ny);
	void tab ();
	void tab (size_t n);

	void setcolor (InkNum color);
	InkNum getcolor ();
	void setbackground (InkNum color);
	InkNum getbackground ();

	void movecharforward ();
	void movecharforward (size_t n);
	void movecharback ();
	void movecharback (size_t n);
	void movecharup ();
	void movecharup (size_t n);
	void movechardown ();
	void movechardown (size_t n);

	void clear_rectangle (int left, int right, int top, int bottom);
	void clear_from_left ();
	void clear_to_right ();
	void clear_from_begin ();
	void clear_to_end ();
	void cls ();

	void deletechar ();

	void textscroll (int nlines= 1, int fromline= 0);
	void textscrollinverse (int nlines= 1, int fromline= 0);
	void scroll (int n);

	void tag_charout (char c);
	void charout (char c);
	void stringout (const std::string & str);
	void do_charout (char c);

	void invertcursor ();

	std::string copychr (BlChar from, BlChar to);

	void tag ();
	void tagoff ();
	bool istagactive () const;
	std::string getkey ();
	std::string inkey ();
private:
	Screen & screen;

	size_t collecting_params;
	std::string params;
	unsigned char actual_control;

	struct DefControlChar {
		size_t nparams;
		bool force;
		void (BlTextWindow::* action) ();
		DefControlChar (size_t nparams, bool force,
				void (BlTextWindow::* action) () ) :
			nparams (nparams),
			force (force),
			action (action)
		{ }
		// Default constructor required by the initialization
		// of the escape map (I don't want to use insert
		// instead of [ ] ).
		DefControlChar ()
		{ }
	};
	static const DefControlChar control [];
	typedef map <char, DefControlChar> escape_t;
	static const escape_t escape;
	static escape_t init_escape ();

	int orgx, orgy, width, height;
	int x, y, savex, savey;
	bool fTag;
	bool cursor_visible;
	bool inversemode;
	bool bright;
	bool underline;
	InkNum pen;
	InkNum paper;
	InkNum realpen;
	InkNum realpaper;
	int maxtcol;
	int maxtrow;

	void forcelegalposition ();
	void set_limits ();
	void set_real_pen ();
	void set_real_paper ();
	void set_real_colors ();

	void ignore ();
	void do_SOH (); // 1
	void do_STX (); // 2
	void do_ETX (); // 3
	void do_ENQ (); // 5
	void do_BEL (); // 7
	void do_BS (); // 8
	void do_TAB (); // 9
	void do_LF (); // 10
	void do_VT (); // 11
	void do_FF (); // 12
	void do_CR (); // 13
	void do_SO (); // 14
	void do_SI (); // 15
	void do_DLE (); // 16
	void do_DC1 (); // 17
	void do_DC2 (); // 18
	void do_DC3 (); // 19
	void do_DC4 (); // 20
	void do_SYN (); // 22
	void do_ETB (); // 23
	void do_CAN (); // 24
	void do_EM (); // 25
	void do_SUB (); // 26
	void do_ESC (); // 27
	void do_FS (); // 28
	void do_RS (); // 30
	void do_US (); // 31

	void do_ESC_A ();
	void do_ESC_B ();
	void do_ESC_C ();
	void do_ESC_D ();
	void do_ESC_E ();
	void do_ESC_H ();
	void do_ESC_I ();
	void do_ESC_J ();
	void do_ESC_K ();
	void do_ESC_L ();
	void do_ESC_M ();
	void do_ESC_N ();
	void do_ESC_Y ();

	void do_ESC_b ();
	void do_ESC_c ();
	void do_ESC_d ();
	void do_ESC_e ();
	void do_ESC_f ();
	void do_ESC_j ();
	void do_ESC_k ();
	void do_ESC_l ();
	void do_ESC_o ();
	void do_ESC_p ();
	void do_ESC_q ();
	void do_ESC_r ();
	void do_ESC_u ();
	void do_ESC_x ();
	void do_ESC_y ();
};


const BlTextWindow::DefControlChar BlTextWindow::control [32]= {
	BlTextWindow::DefControlChar (0, false, & BlTextWindow::ignore), // NUL
	BlTextWindow::DefControlChar (1, true,  & BlTextWindow::do_SOH), // SOH
	BlTextWindow::DefControlChar (0, false, & BlTextWindow::do_STX), // STX
	BlTextWindow::DefControlChar (0, false, & BlTextWindow::do_ETX), // ETX
	BlTextWindow::DefControlChar (1, false, & BlTextWindow::ignore), // EOT
	BlTextWindow::DefControlChar (1, false, & BlTextWindow::do_ENQ), // ENQ
	BlTextWindow::DefControlChar (0, false, & BlTextWindow::ignore), // ACK
	BlTextWindow::DefControlChar (0, false, & BlTextWindow::do_BEL), // BEL
	BlTextWindow::DefControlChar (0, true,  & BlTextWindow::do_BS ), // BS
	BlTextWindow::DefControlChar (0, true,  & BlTextWindow::do_TAB), // TAB
	BlTextWindow::DefControlChar (0, true,  & BlTextWindow::do_LF ), // LF
	BlTextWindow::DefControlChar (0, true,  & BlTextWindow::do_VT ), // VT
	BlTextWindow::DefControlChar (0, false, & BlTextWindow::do_FF ), // FF
	BlTextWindow::DefControlChar (0, true,  & BlTextWindow::do_CR ), // CR
	BlTextWindow::DefControlChar (1, false, & BlTextWindow::do_SO ), // SO
	BlTextWindow::DefControlChar (1, false, & BlTextWindow::do_SI ), // SI
	BlTextWindow::DefControlChar (0, true,  & BlTextWindow::do_DLE), // DLE
	BlTextWindow::DefControlChar (0, true,  & BlTextWindow::do_DC1), // DC1
	BlTextWindow::DefControlChar (0, true,  & BlTextWindow::do_DC2), // DC2
	BlTextWindow::DefControlChar (0, true,  & BlTextWindow::do_DC3), // DC3
	BlTextWindow::DefControlChar (0, true,  & BlTextWindow::do_DC4), // DC4
	BlTextWindow::DefControlChar (0, false, & BlTextWindow::ignore), // NAK
	BlTextWindow::DefControlChar (1, false, & BlTextWindow::do_SYN), // SYN
	BlTextWindow::DefControlChar (1, false, & BlTextWindow::do_ETB), // ETB
	BlTextWindow::DefControlChar (0, false, & BlTextWindow::do_CAN), // CAN
	BlTextWindow::DefControlChar (9, false, & BlTextWindow::do_EM ), // EM
	BlTextWindow::DefControlChar (4, false, & BlTextWindow::ignore), // SUB
	BlTextWindow::DefControlChar (1, false, & BlTextWindow::do_ESC), // ESC
	BlTextWindow::DefControlChar (3, false, & BlTextWindow::do_FS ), // FS
	BlTextWindow::DefControlChar (2, false, & BlTextWindow::ignore), // GS
	BlTextWindow::DefControlChar (0, false, & BlTextWindow::do_RS ), // RS
	BlTextWindow::DefControlChar (2, false, & BlTextWindow::do_US ), // US
};

BlTextWindow::escape_t BlTextWindow::init_escape ()
{
	escape_t aux;
	DefControlChar ignore0 (0, false, & BlTextWindow::ignore);
	DefControlChar ignore1 (1, false, & BlTextWindow::ignore);
	aux ['0']= ignore0; // Status line inactive.
	aux ['1']= ignore0; // Status line active.
	aux ['2']= ignore1; // Select national character set
	aux ['3']= ignore1; // Set mode
	aux ['A']= DefControlChar (0, true,  & BlTextWindow::do_ESC_A);
	aux ['B']= DefControlChar (0, true,  & BlTextWindow::do_ESC_B);
	aux ['C']= DefControlChar (0, true,  & BlTextWindow::do_ESC_C);
	aux ['D']= DefControlChar (0, true,  & BlTextWindow::do_ESC_D);
	aux ['E']= DefControlChar (0, false, & BlTextWindow::do_ESC_E);
	aux ['H']= DefControlChar (0, false, & BlTextWindow::do_ESC_H);
	aux ['I']= DefControlChar (0, true,  & BlTextWindow::do_ESC_I);
	aux ['J']= DefControlChar (0, true,  & BlTextWindow::do_ESC_J);
	aux ['K']= DefControlChar (0, true,  & BlTextWindow::do_ESC_K);
	aux ['L']= DefControlChar (0, true,  & BlTextWindow::do_ESC_L);
	aux ['M']= DefControlChar (0, true,  & BlTextWindow::do_ESC_M);
	aux ['N']= DefControlChar (0, true,  & BlTextWindow::do_ESC_N);
	aux ['Y']= DefControlChar (2, false, & BlTextWindow::do_ESC_Y);
	aux ['b']= DefControlChar (1, false, & BlTextWindow::do_ESC_b);
	aux ['c']= DefControlChar (1, false, & BlTextWindow::do_ESC_c);
	aux ['d']= DefControlChar (0, true,  & BlTextWindow::do_ESC_d);
	aux ['e']= DefControlChar (0, false, & BlTextWindow::do_ESC_e);
	aux ['f']= DefControlChar (0, false, & BlTextWindow::do_ESC_f);
	aux ['j']= DefControlChar (0, false, & BlTextWindow::do_ESC_j);
	aux ['k']= DefControlChar (0, false, & BlTextWindow::do_ESC_k);
	aux ['l']= DefControlChar (0, true,  & BlTextWindow::do_ESC_l);
	aux ['o']= DefControlChar (0, true,  & BlTextWindow::do_ESC_o);
	aux ['p']= DefControlChar (0, false, & BlTextWindow::do_ESC_p);
	aux ['q']= DefControlChar (0, false, & BlTextWindow::do_ESC_q);
	aux ['r']= DefControlChar (0, false, & BlTextWindow::do_ESC_r);
	aux ['u']= DefControlChar (0, false, & BlTextWindow::do_ESC_u);
	aux ['x']= DefControlChar (0, false, & BlTextWindow::do_ESC_x);
	aux ['y']= DefControlChar (0, false, & BlTextWindow::do_ESC_y);
	return aux;
}

const BlTextWindow::escape_t BlTextWindow::escape=
	BlTextWindow::init_escape ();

BlTextWindow::BlTextWindow (Screen & sc) :
	screen (sc),
	collecting_params (0),
	fTag (false),
	cursor_visible (true),
	inversemode (false),
	bright (false),
	underline (false)
{
	TRF;

	set_limits ();
	setdefault ();
	defaultcolors ();
}

BlTextWindow::BlTextWindow (Screen & sc, int x1, int x2, int y1, int y2) :
	screen (sc),
	collecting_params (0),
	fTag (false),
	cursor_visible (true),
	inversemode (false),
	bright (false),
	underline (false)
{
	TRF;

	set_limits ();
	set (x1, x2, y1, y2);
	defaultcolors ();
}

Screen * BlTextWindow::get_screen ()
{
	return & screen;
}

void BlTextWindow::set_limits ()
{
	maxtcol= screen.get_maxtcol ();
	maxtrow= screen.get_maxtrow ();
}

void BlTextWindow::set (int x1, int x2, int y1, int y2)
{
	TRF;

	#if 0
	if (x1 == -1 && x2 == -1)
	{
		x1= 0;
		x2= maxtcol - 1;
	}
	else
	{
		if (x1 > x2) swap (x1, x2);
		if (x1 >= maxtcol) x1= maxtcol - 1;
		if (x2 >= maxtcol) x2= maxtcol - 1;
	}
	if (y1 == -1 && y2 == -1)
	{
		y1= 0;
		y2= maxtrow - 1;
	}
	else
	{
		if (y1 > y2) swap (y1, y2);
		if (y1 >= maxtrow) y1= maxtrow - 1;
		if (y2 >= maxtrow) y2= maxtrow - 1;
	}
	#endif

	if (x1 == -1)
		x1= 0;
	if (x2 == -1)
		x2= maxtcol - 1;
	if (y1 == -1)
		y1= 0;
	if (y2 == -1)
		y2= maxtrow - 1;
	if (y1 > y2) swap (y1, y2);
	if (y1 >= maxtrow) y1= maxtrow - 1;
	if (y2 >= maxtrow) y2= maxtrow - 1;

	orgx= x1; orgy= y1;
	width= x2 - x1 + 1;
	height= y2 - y1 + 1;
	x= y= 0;
}

void BlTextWindow::setdefault ()
{
	TRF;

	set (0, maxtcol - 1, 0, maxtrow - 1);
}

void BlTextWindow::set_real_pen ()
{
	realpen= pen;
	if (bright)
	{
		if (realpen < 8)
			realpen+= 8;
	}
}

void BlTextWindow::set_real_paper ()
{
	realpaper= paper;
	if (bright)
	{
		if (realpaper < 8)
			realpaper+= 8;
	}
}

void BlTextWindow::set_real_colors ()
{
	set_real_pen ();
	set_real_paper ();
}

void BlTextWindow::defaultcolors ()
{
	pen= screen.get_default_pen ();
	paper= screen.get_default_paper ();
	set_real_colors ();
}

void BlTextWindow::synchronize_suspend ()
{
	screen.suspend_synchro ();
}

void BlTextWindow::synchronize_restart ()
{
	screen.restart_synchro ();
}

void BlTextWindow::setbright (bool active)
{
	bool previous= bright;
	bright= active;
	if (bright != previous)
	{
		set_real_colors ();
	}
}

void BlTextWindow::setinverse (bool active)
{
	inversemode= active;
}

void BlTextWindow::gotoxy (int nx, int ny)
{
	TRFDEBS ("x: " << x << " y: " << y);

	x= nx;
	y= ny;
}

void BlTextWindow::forcelegalposition ()
{
	if (x >= width)
		{ x= 0; ++y; }
	if (x < 0)
		{ x= width - 1; --y; }
	if (y < 0)
	{
		y= 0;
		textscrollinverse ();
	}
	if (y >= height)
	{
		textscroll ();
		y= height - 1;
	}
}

void BlTextWindow::tab ()
{
	forcelegalposition ();
	int zone= sysvar::get16 (sysvar::Zone);
	if (zone == 0)
		zone= 8;
	if (x >= (width / zone) * zone)
	{
		//cerr << "Fin de linea" << endl;
		int yy= orgy + y;
		for ( ; x < width; ++x)
		{
			screen.print (' ', orgx + x, yy, pen, paper,
				inversemode, underline);
		}
		x= 0;
		++y;
	}
	else
	{
		int yy= orgy + y;
		do {
			screen.print (' ', orgx + x, yy, pen, paper,
				inversemode, underline);
			++x;
		} while (x % zone);
	}
}

void BlTextWindow::tab (size_t n)
{
	forcelegalposition ();
	int col= n;
	if (x > col)
	{
		do {
			charout (' ');
		} while (x < width);
	}
	int maxpos= std::min (col, width);
	while (x < maxpos)
		charout (' ');
	x= col;
}

void BlTextWindow::setcolor (InkNum color)
{
	pen= color;
	set_real_pen ();
}

InkNum BlTextWindow::getcolor ()
{
	return pen;
}

void BlTextWindow::setbackground (InkNum color)
{
	paper= color;
	set_real_paper ();
}

InkNum BlTextWindow::getbackground ()
{
	return paper;
}

void BlTextWindow::movecharforward ()
{
	forcelegalposition ();
	++x;
}

void BlTextWindow::movecharforward (size_t n)
{
	for ( ; n > 0; --n)
		movecharforward ();
}

void BlTextWindow::movecharback ()
{
	forcelegalposition ();
	--x;
}

void BlTextWindow::movecharback (size_t n)
{
	for ( ; n > 0; --n)
		movecharback ();
}

void BlTextWindow::movecharup ()
{
	forcelegalposition ();
	--y;
}

void BlTextWindow::movecharup (size_t n)
{
	for ( ; n > 0; --n)
		movecharup ();
}

void BlTextWindow::movechardown ()
{
	forcelegalposition ();
	++y;
}

void BlTextWindow::movechardown (size_t n)
{
	for ( ; n > 0; --n)
		movechardown ();
}

void BlTextWindow::clear_rectangle (int left, int right, int top, int bottom)
{
	TRF;

	int charwidth= screen.getcharwidth ();
	int charheight= screen.getcharheight ();

	int x1= (orgx + left) * charwidth;
	int x2= (orgx + right + 1) * charwidth - 1;
	int y1= (orgy + top) * charheight;
	int y2= (orgy + bottom + 1) * charheight - 1;

	screen.setactiveink (paper);
	screen.do_fill_rectangle (x1, y1, x2, y2, false);
}

void BlTextWindow::clear_from_left ()
{
	clear_rectangle (0, x, y, y);
}

void BlTextWindow::clear_to_right ()
{
	clear_rectangle (x, width - 1, y, y);
}

void BlTextWindow::clear_from_begin ()
{
	if (y > 0)
		clear_rectangle (0, width - 1, 0, y - 1);
	clear_from_left ();
}

void BlTextWindow::clear_to_end ()
{
	clear_to_right ();
	if (y < height - 1)
		clear_rectangle (0, width - 1, y + 1, height - 1);
}

void BlTextWindow::cls ()
{
	TRF;

	x= y= 0;
	clear_rectangle (0, width - 1, 0, height - 1);
}

void BlTextWindow::deletechar ()
{
	int charwidth= screen.getcharwidth ();
	int charheight= screen.getcharheight ();

	int x1= (orgx + x) * charwidth;
	int y1= (orgy + y) * charheight;
	int x2= (orgx + width) * charwidth - 1;
	int y2= y1 + charheight - 1;

	screen.do_scroll_left (x1, y1, x2, y2, charwidth, paper);
}

void BlTextWindow::textscroll (int nlines, int fromline)
{
	TRF;

	int charwidth= screen.getcharwidth ();
	int charheight= screen.getcharheight ();
	int amount= nlines * charheight;

	int x1= orgx * charwidth;
	int y1= (orgy + fromline) * charheight;
	int x2= (orgx + width) * charwidth - 1;
	int y2= (orgy + height) * charheight - 1;

	screen.do_scroll_up (x1, y1, x2, y2, amount, paper);
}

void BlTextWindow::textscrollinverse (int nlines, int fromline)
{
	TRF;

	int charwidth= screen.getcharwidth ();
	int charheight= screen.getcharheight ();

	int amount= nlines * charheight;

	int x1= orgx * charwidth;
	int x2= (orgx + width) * charwidth - 1;
	int y1= (orgy + fromline) * charheight;
	int y2= (orgy + height) * charheight - 1;

	screen.do_scroll_down (x1, y1, x2, y2, amount, paper);
}

void BlTextWindow::scroll (int n)
{
	TRFDEBS (n);

	forcelegalposition ();
	if (n < 0)
	{
		textscrollinverse (-n);
		gotoxy (0, 0);
	}
	else
	{
		textscroll (n);
		gotoxy (0, height - 1);
	}
	screen.flush ();
}

void BlTextWindow::tag_charout (char c)
{
	TRF;

	screen.printxy (c, true, false, false);
}

void BlTextWindow::charout (char c)
{
	TRF;

	if (fTag)
	{
		tag_charout (c);
		return;
	}
	if (collecting_params)
	{
		TRDEB ("collecting params");
		params+= c;
		if (--collecting_params == 0)
		{
			const DefControlChar & defcontrol=
				actual_control < 32 ?
				control [actual_control] :
				escape.find (actual_control)->second;
			if (defcontrol.force)
				forcelegalposition ();
			(this->* defcontrol.action) ();
		}
		return;
	}
	if (iscontrolchar (c) )
	{
		TRDEB ("Is control char");
		actual_control= c;
		params.erase ();
		const DefControlChar & defcontrol=
			control [actual_control];
		if (defcontrol.nparams > 0)
		{
			TRDEB ("Init collecting");
			collecting_params= defcontrol.nparams;
		}
		else
		{
			if (defcontrol.force)
				forcelegalposition ();
			(this->* defcontrol.action) ();
		}
		return;
	}
	forcelegalposition ();

	screen.print (c, orgx + x, orgy + y, pen, paper,
		inversemode, underline);
	++x;
}

void BlTextWindow::stringout (const std::string & str)
{
	for (std::string::size_type i= 0, l= str.size (); i < l; ++i)
		charout (str [i] );
	screen.flush ();
}

void BlTextWindow::do_charout (char c)
{
	screen.print (c, orgx + x, orgy + y, pen, paper,
		inversemode, underline);
	++x;
}

void BlTextWindow::invertcursor ()
{
	forcelegalposition ();
	if (! cursor_visible)
		return;

	int charwidth= screen.getcharwidth ();
	int charheight= screen.getcharheight ();

	int x1= (orgx + x) * charwidth;
	int y1= (orgy + y) * charheight;
	int x1ini= x1;
	int y1ini= y1 + charheight - 2;
	int x1end= x1 + charwidth;
	int y1end= y1 + charheight;

	screen.invert_rectangle (x1ini, y1ini, x1end, y1end);
}

std::string BlTextWindow::copychr (BlChar from, BlChar to)
{
	// I don't tested yet if that is done in the cpc
	forcelegalposition ();

	int x1= (orgx + x) * screen.getcharwidth ();
	int y1= (orgy + y) * screen.getcharheight ();
	//return copychrat (x1, y1, from , to);
	return screen.copychrat (x1, y1, from, to);
}

void BlTextWindow::tag ()
{
	fTag= true;
}

void BlTextWindow::tagoff ()
{
	fTag= false;
}

bool BlTextWindow::istagactive () const
{
	return fTag;
}

std::string BlTextWindow::getkey ()
{
	TRF;

	#if 1

	return screen.getkey ();

	#else

	QueueKey & queuekey= screen.getQueueKey ();

	#ifdef BLASSIC_USE_X

	while (queuekey.empty () )
	{
		//wait_X_event ();
		screen.wait_event ();
		if (fInterrupted)
			return std::string ();
	}

	return queuekey.pop ();

	#else

	for (;;)
	{
		blassic::idle ();
		if (! queuekey.empty () )
			return queuekey.pop ();
	}

	#endif

	#endif
}

std::string BlTextWindow::inkey ()
{
	TRF;

	#if 1

	return screen.inkey ();

	#else

	QueueKey & queuekey= screen.getQueueKey ();

	blassic::idle ();
	if (queuekey.empty () )
		return std::string ();

	return queuekey.pop ();

	#endif
}

void BlTextWindow::ignore ()
{
	/* Nothing to do */
}

void BlTextWindow::do_SOH () // 1
{
	do_charout (params [0] );
}

void BlTextWindow::do_STX () // 2
{
	cursor_visible= false;
}

void BlTextWindow::do_ETX () // 3
{
	cursor_visible= true;
}

void BlTextWindow::do_ENQ () // 5
{
	tag_charout (params [0] );
}

void BlTextWindow::do_BEL () // 7
{
	//blassic::graphics::ring ();
	screen.ring_my_bell ();
}

void BlTextWindow::do_BS () // 8
{
	--x;
}

void BlTextWindow::do_TAB () // 9
{
	++x;
}

void BlTextWindow::do_LF () // 10
{
	//x= 0;
	++y;
}

void BlTextWindow::do_VT () // 11
{
	--y;
}

void BlTextWindow::do_FF () // 12
{
	cls ();
}

void BlTextWindow::do_CR () // 13
{
	x= 0;
}

void BlTextWindow::do_SO () // 14
{
	setbackground (static_cast <unsigned char> (params [0] ) % 16);
}

void BlTextWindow::do_SI () // 15
{
	setcolor (static_cast <unsigned char> (params [0] ) % 16 );
}

void BlTextWindow::do_DLE () // 16
{
	clear_rectangle (x, x, y, y);
}

void BlTextWindow::do_DC1 () // 17
{
	clear_from_left ();
}

void BlTextWindow::do_DC2 () // 18
{
	clear_to_right ();
}

void BlTextWindow::do_DC3 () // 19
{
	clear_from_begin ();
}

void BlTextWindow::do_DC4 () // 20
{
	clear_to_end ();
}

void BlTextWindow::do_SYN () // 22
{
	blassic::graphics::settransparent
		(static_cast <unsigned char> (params [0] ) );
}

void BlTextWindow::do_ETB () // 23
{
	// Ink mode, only CPC modes are allowed.
	blassic::graphics::setdrawmode
		(static_cast <unsigned char> (params [0] ) % 4);
}

void BlTextWindow::do_CAN () // 24
{
	//swap (foreground, background);
	swap (pen, paper);
	swap (realpen, realpaper);
}

void BlTextWindow::do_EM () // 25
{
	int symbol= static_cast <unsigned char> (params [0] );
	// Avoid generate an error if out of range.
	if (symbol < symbol_after_is)
		return;
	unsigned char byte [8];
	params.copy (reinterpret_cast <char *> (& byte [0] ), 8, 1);
	blassic::graphics::definesymbol (symbol, byte);
}

void BlTextWindow::do_SUB () // 26
{
	set (static_cast <unsigned char> (params [0] ),
		static_cast <unsigned char> (params [1] ),
		static_cast <unsigned char> (params [2] ),
		static_cast <unsigned char> (params [3] ) );
}

void BlTextWindow::do_ESC () // 27
{
	char c= params [0];
	escape_t::const_iterator it= escape.find (c);
	if (it == escape.end () )
	{
		do_charout (c);
		return;
	}
	const DefControlChar & defcontrol= it->second;
	if (defcontrol.nparams == 0)
	{
		if (defcontrol.force)
			forcelegalposition ();
		(this->* defcontrol.action) ();
		return;
	}
	collecting_params= defcontrol.nparams;
	// All escapes used have non-control codes,
	// then we can use the character code
	// without confusion.
	ASSERT (! iscontrolchar (c) );

	actual_control= c;
	params.erase ();
}

void BlTextWindow::do_FS () // 28
{
	int ink= static_cast <unsigned char> (params [0] ) % 16;
	int color= static_cast <unsigned char> (params [1] ) % 32;
	if (color > 26)
		return;
	// Third parameter (flashing ink) ignored.
	blassic::graphics::ink (ink, color);
}

void BlTextWindow::do_RS () // 30
{
	x= 0; y= 0;
}

void BlTextWindow::do_US () // 31
{
	int xx= static_cast <unsigned char> (params [0] );
	int yy= static_cast <unsigned char> (params [1] );
	if (xx == 0 || yy == 0)
		return;
	x= xx - 1; y= yy - 1;
}

void BlTextWindow::do_ESC_A ()
{
	if (y > 0)
		--y;
}

void BlTextWindow::do_ESC_B ()
{
	if (y < height - 1)
		++y;
}

void BlTextWindow::do_ESC_C ()
{
	if (x < width - 1)
		++x;
}

void BlTextWindow::do_ESC_D ()
{
	if (x > 0)
		--x;
}

void BlTextWindow::do_ESC_E ()
{
	clear_rectangle (0, width - 1, 0, height - 1);
}

void BlTextWindow::do_ESC_H ()
{
	x= 0; y= 0;
}

void BlTextWindow::do_ESC_I ()
{
	--y;
}

void BlTextWindow::do_ESC_J ()
{
	clear_to_end ();
}

void BlTextWindow::do_ESC_K ()
{
	clear_to_right ();
}

void BlTextWindow::do_ESC_L ()
{
	textscrollinverse (0, y);
}

void BlTextWindow::do_ESC_M ()
{
	textscroll (0, y);
}

void BlTextWindow::do_ESC_N ()
{
	deletechar ();
}

void BlTextWindow::do_ESC_Y ()
{
	y= static_cast <unsigned char> (params [0] ) - 32;
	x= static_cast <unsigned char> (params [1] ) - 32;
	if (y > height - 1)
		y= height - 1;
	if (x > width - 1)
		x= width - 1;
}

void BlTextWindow::do_ESC_b ()
{
	// Testing
	setcolor (static_cast <unsigned char> (params [0] ) & 0x3F);
}

void BlTextWindow::do_ESC_c ()
{
	// Testing
	setbackground (static_cast <unsigned char> (params [0] ) & 0x3F);
}

void BlTextWindow::do_ESC_d ()
{
	clear_from_begin ();
}

void BlTextWindow::do_ESC_e ()
{
	cursor_visible= true;
}

void BlTextWindow::do_ESC_f ()
{
	cursor_visible= false;
}

void BlTextWindow::do_ESC_j ()
{
	savex= x; savey= y;
}

void BlTextWindow::do_ESC_k ()
{
	x= savex; y= savey;
}

void BlTextWindow::do_ESC_l ()
{
	clear_rectangle (0, width - 1, y, y);
}

void BlTextWindow::do_ESC_o ()
{
	clear_from_left ();
}

void BlTextWindow::do_ESC_p ()
{
	inversemode= true;
	//swap (foreground, background);
}

void BlTextWindow::do_ESC_q ()
{
	inversemode= false;
}

void BlTextWindow::do_ESC_r ()
{
	TRF;
	underline= true;
}

void BlTextWindow::do_ESC_u ()
{
	TRF;
	underline= false;
}

void BlTextWindow::do_ESC_x ()
{
	// set mode 24 x 80
	setdefault ();
	cls ();
	set (0, 79, 0, 23);
}

void BlTextWindow::do_ESC_y ()
{
	// unset mode 24 x 80
	setdefault ();
	cls ();
}



// Map encapsultated to check access. Windows must be accessed only
// throug file, then accesing a window that not exists is a internal
// error. Also destruction of windows is now granted.

class MapWindow {
	typedef map <BlChannel, BlTextWindow *> map_type;
	map_type mw;
public:
	typedef map_type::iterator iterator;
	typedef map_type::const_iterator const_iterator;
	typedef map_type::value_type value_type;
	typedef map_type::key_type key_type;
	typedef map_type::mapped_type mapped_type;

	~MapWindow ();
	iterator begin ();
	iterator end ();
	const_iterator begin () const;
	const_iterator end () const;
	// operators [] are checked, fail if the key not exist.
	mapped_type operator [] (key_type n) const;
	// get return NULL if not exist.
	mapped_type get (key_type n) const;
	iterator find (key_type n);
	static void killwindowifnotzero (const value_type & vt);
	void clear ();
	void insert (const value_type & val);
	void erase (iterator pos);
};

MapWindow::~MapWindow ()
{
	// Ensure window destruction on exit.
	clear ();
}

MapWindow::iterator MapWindow::begin ()
{
	return mw.begin ();
}

MapWindow::iterator MapWindow::end ()
{
	return mw.end ();
}

MapWindow::const_iterator MapWindow::begin () const
{
	return mw.begin ();
}

MapWindow::const_iterator MapWindow::end () const
{
	return mw.end ();
}

MapWindow::mapped_type MapWindow::operator [] (key_type n) const
{
	TRF;

	const_iterator it= mw.find (n);
	if (it == end () )
	{
		TRDEB ("No such window");
		if (showdebuginfo () )
			cerr << "Trying to access to window " <<
				n << " that not exists" << endl;
		throw ErrBlassicInternal;
	}
	return it->second;
}

MapWindow::mapped_type MapWindow::get (key_type n) const
{
	const_iterator it= mw.find (n);
	if (it == end () )
		return NULL;
	else
		return it->second;
}

MapWindow::iterator MapWindow::find (key_type n)
{
	return mw.find (n);
}

void MapWindow::killwindowifnotzero (const value_type & vt)
{
	// Probando
	//if (vt.first != BlChannel (0) )
		delete vt.second;
}

void MapWindow::clear ()
{
	for_each (begin (), end (), & MapWindow::killwindowifnotzero);
	mw.clear ();
}

void MapWindow::insert (const value_type & val)
{
	// Fail if already exist.
	std::pair <iterator, bool> r= mw.insert (val);
	if (! r.second)
	{
		if (showdebuginfo () )
			cerr << "Trying to create window " <<
				val.first << " that already exists" <<
				endl;
		throw ErrBlassicInternal;
	}
}

void MapWindow::erase (iterator pos)
{
	// Unchecked, actually.
	mw.erase (pos);
}


MapWindow mapwindow;


void destroy_text_windows ()
{
	mapwindow.clear ();
}

#if 0
void recreate_windows ()
{
	TRACEFUNC (tr, "recreate_windows");

	#if 0
	windowzero.setdefault ();
	windowzero.defaultcolors ();
	windowzero.cls ();

	//std::for_each (mapwindow.begin (), mapwindow.end (),
	//	killwindowifnotzero);
	//mapwindow.clear ();
	destroy_text_windows ();

	//mapwindow [0]= & windowzero;
	mapwindow.insert (make_pair (BlChannel (0), & windowzero) );

	#else

	BlTextWindow * pwinzero= mapwindow.get (0);
	destroy_text_windows ();
	if (pwinzero)
	{
		mapwindow.insert (make_pair (BlChannel (0), pwinzero) );
		pwinzero->setdefault ();
		pwinzero->defaultcolors ();
		pwinzero->cls ();
	}

	#endif
}
#endif


typedef map <std::string, Image *> imagemap_t;
imagemap_t imagemap;


} // namespace graphics_impl


char skipblank (const char * & s)
{
	char c;
	while ( (c= * s) == ' ' || c == '\t')
		++s;
	return c;
}

int getnum (const char * & s)
{
	int r= 0;
	char c;
	if (! isdigit (* s) )
	{
		if (showdebuginfo () )
			cerr << "Inavlid number in DRAW string" << endl;
		throw ErrSyntax;
	}
	while ( isdigit (c= * s) )
	{
		r*= 10;
		r+= c - '0';
		++s;
	}
	//cerr << r;
	return r;
}

void drawsyntaxerror ()
{
	if (showdebuginfo () )
		cerr << "Invalid syntax in DRAW string" << endl;
	throw ErrSyntax;
}

enum TypeMove { MoveAbs, MovePos, MoveNeg };

TypeMove gettypemove (const char * & s)
{
	TypeMove typemove= MoveAbs;
	char c= skipblank (s);
	switch (c)
	{
	case '+':
		//cerr << '+';
		typemove= MovePos;
		++s;
		skipblank (s);
		break;
	case '-':
		//cerr << '-';
		typemove= MoveNeg;
		++s;
		skipblank (s);
		break;
	}
	return typemove;
}

inline void adjust (int & value, TypeMove t, int last)
{
	switch (t)
	{
	case MoveAbs:
		break;
	case MovePos:
		value= last + value;
		break;
	case MoveNeg:
		value= last - value;
		break;
	}
}


namespace graphics {

using namespace graphics_impl;

}

} // namespace blassic


#endif
// BLASSIC_HAS_GRAPHICS


void blassic::graphics::draw (const std::string & str)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	Screen & sc= * getcurrentscreen ();

	const char * s= str.c_str ();
	char c;
	int i;
	while ( (c= skipblank (s) ) != '\0')
	{
		//cerr << c;
		++s;
		bool nopaint= false;
		if (c == 'B')
		{
			c= skipblank (s);
			//cerr << c;
			++s;
			nopaint= true;
		}
		switch (c)
		{
		case 'M':
			{
				TypeMove mx= gettypemove (s);
				int x= getnum (s);
				adjust (x, mx, sc.get_lastx () );

				c= skipblank (s);
				if (c != ',')
				{
					//throw ErrSyntax;
					drawsyntaxerror ();
				}
				//cerr << ',';
				++s;
				TypeMove my= gettypemove (s);
				int y= getnum (s);
				adjust (y, my, sc.get_lasty () );

				if (nopaint)
					sc.do_move (x, y);
				else
					sc.do_line (x, y);
			}
			break;
		case 'U':
			skipblank (s);
			i= getnum (s);
			if (nopaint)
			{
				//blassic::graphics::move (lastx, lasty - i);
				sc.do_move_rel (0, -i);
			}
			else
			{
				//blassic::graphics::line (lastx, lasty - i);
				sc.do_line_rel (0, -i);
			}
			break;
		case 'D':
			skipblank (s);
			i= getnum (s);
			if (nopaint)
			{
				//blassic::graphics::move (lastx, lasty + i);
				sc.do_move_rel (0, i);
			}
			else
			{
				//blassic::graphics::line (lastx, lasty + i);
				sc.do_line_rel (0, i);
			}
			break;
		case 'R':
			skipblank (s);
			i= getnum (s);
			if (nopaint)
			{
				//blassic::graphics::move (lastx + i, lasty);
				sc.do_move_rel (i, 0);
			}
			else
			{
				//blassic::graphics::line (lastx + i, lasty);
				sc.do_line_rel (i, 0);
			}
			break;
		case 'L':
			skipblank (s);
			i= getnum (s);
			if (nopaint)
			{
				//blassic::graphics::move (lastx - i, lasty);
				sc.do_move_rel (- i, 0);
			}
			else
			{
				//blassic::graphics::line (lastx - i, lasty);
				sc.do_line_rel (- i, 0);
			}
			break;
		case 'E':
			skipblank (s);
			i= getnum (s);
			if (nopaint)
			{
				//blassic::graphics::move (lastx + i, lasty - i);
				sc.do_move_rel (i, - i);
			}
			else
			{
				//blassic::graphics::line (lastx + i, lasty - i);
				sc.do_line_rel (i, - i);
			}
			break;
		case 'F':
			skipblank (s);
			i= getnum (s);
			if (nopaint)
			{
				//blassic::graphics::move (lastx + i, lasty + i);
				sc.do_move_rel (i, i);
			}
			else
			{
				//blassic::graphics::line (lastx + i, lasty + i);
				sc.do_line_rel (i, i);
			}
			break;
		case 'G':
			skipblank (s);
			i= getnum (s);
			if (nopaint)
			{
				//blassic::graphics::move (lastx - i, lasty + i);
				sc.do_move_rel (- i, i);
			}
			else
			{
				//blassic::graphics::line (lastx - i, lasty + i);
				sc.do_line_rel (- i, i);
			}
			break;
		case 'H':
			skipblank (s);
			i= getnum (s);
			if (nopaint)
			{
				//blassic::graphics::move (lastx - i, lasty - i);
				sc.do_move_rel (- i, - i);
			}
			else
			{
				//blassic::graphics::line (lastx - i, lasty - i);
				sc.do_line_rel (- i, - i);
			}
			break;
		case 'C':
			c= skipblank (s);
			if (! isdigit (c) )
				//throw ErrSyntax;
				drawsyntaxerror ();
			//blassic::graphics::setcolor (c - '0');
			sc.set_graphics_pen (c - '0');
			++s;
			break;
		case 'X':
			{
				std::string x;
				while ( (c= *s) != ';' && c != '\0')
				{
					x+= c;
					++s;
				}
				if (c != ';')
					//throw ErrSyntax;
					drawsyntaxerror ();
				++s;
				if (typeofvar (x) != VarString)
					throw ErrMismatch;
				std::string xx= evaluatevarstring (x);
				draw (xx);
			}
			break;
		case ';':
			break;
		default:
			//throw ErrSyntax;
			drawsyntaxerror ();
		}
	}
	//cerr << endl;

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (str);

	#endif
}

blassic::graphics::Point blassic::graphics::getlast ()
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	Screen * pscreen= getcurrentscreen ();

	return Point (pscreen->get_lastx (), pscreen->get_lasty () );

	#else

	return Point (0, 0);

	#endif
}

std::string blassic::graphics::inkey ()
{
	TRF;

	#ifdef  BLASSIC_HAS_GRAPHICS

	#if 1

	return getcurrentscreen ()->inkey ();

	#else

	QueueKey & queuekey= getcurrentscreen ()->getQueueKey ();

	blassic::idle ();
	if (queuekey.empty () )
		return std::string ();

	return queuekey.pop ();

	#endif

	#else
	// No BLASSIC_HAS_GRAPHICS

	ASSERT (false);
	throw ErrBlassicInternal; // Make the compiler happy.

	#endif
}

bool blassic::graphics::pollin ()
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	QueueKey & queuekey= getcurrentscreen ()->getQueueKey ();

	blassic::idle ();
	return ! queuekey.empty ();

	#else

	throw ErrBlassicInternal; // Make the compiler happy.

	#endif
}

bool blassic::graphics::pollin (BlChannel ch)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	//QueueKey & queuekey= getcurrentscreen ()->getQueueKey ();
	QueueKey & queuekey= mapwindow [ch]->get_screen ()->getQueueKey ();

	blassic::idle ();
	return ! queuekey.empty ();

	#else

	touch (ch);
	throw ErrBlassicInternal;

	#endif
}

std::string blassic::graphics::getkey ()
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	#if 1

	return getcurrentscreen ()->getkey ();

	#else

	Screen & screen= * getcurrentscreen ();
	QueueKey & queuekey= screen.getQueueKey ();

	#ifdef BLASSIC_USE_X

	while (queuekey.empty () )
	{
		//wait_X_event ();
		screen.wait_event ();
	}

	return queuekey.pop ();

	#else

	for (;;)
	{
		blassic::idle ();
		if (! queuekey.empty () )
			return queuekey.pop ();

		#if 0

		// Reduces cpu usage.
		#ifdef BLASSIC_USE_WINDOWS_GRAPHICS
		Sleep (0);
		#endif

		#endif
	}

	#endif

	#endif

	#else
	// No BLASSIC_HAS_GRAPHICS

	ASSERT (false); // Must not came here.
	throw ErrBlassicInternal; // Make the compiler happy.

	#endif
}

std::string blassic::graphics::getkey (BlChannel ch)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	return mapwindow [ch]->getkey ();

	#else

	touch (ch);
	throw ErrBlassicInternal;

	#endif
}

std::string blassic::graphics::inkey (BlChannel ch)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	return mapwindow [ch]->inkey ();

	#else

	touch (ch);
	throw ErrBlassicInternal;

	#endif
}

int blassic::graphics::keypressed (int keynum)
{
	TRF;

	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	#if 0

	if (keynum < 0 || keynum > static_cast <int> (MAXINKEYCODE) )
		return -1;
	blassic::idle ();
	if (! keypressedmap [presscode [keynum] ] )
		return -1;
	int r= 0;
	const int shiftpressed= 32, ctrlpressed= 128;

	#ifdef BLASSIC_USE_X
	if (keypressedmap [XK_Shift_L] || keypressedmap [XK_Shift_R] )
		r|= shiftpressed;
	if (keypressedmap [XK_Control_L] || keypressedmap [XK_Control_R] )
		r|= ctrlpressed;
	#endif

	return r;

	#else

	//return ::keypressed (keynum);
	return getcurrentscreen ()->keypressed (keynum);

	#endif

	#else
	// No graphics.

	touch (keynum);
	ASSERT (false); // Must not came here.
	throw ErrBlassicInternal; // Make the compiler happy.

	#endif
}

void blassic::graphics::setcolor (BlChannel ch, int color)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->setcolor (color);

	#else

	touch (ch, color);
	no_graphics_support ();

	#endif
}

int blassic::graphics::getcolor (BlChannel ch)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return mapwindow [ch]->getcolor ();

	#else

	touch (ch);
	no_graphics_support ();
	throw ErrBlassicInternal; // Make the compiler happy

	#endif
}

void blassic::graphics::setbackground (BlChannel ch, int color)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->setbackground (color);

	#else

	touch (ch, color);
	no_graphics_support ();

	#endif
}

int blassic::graphics::getbackground (BlChannel ch)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return mapwindow [ch]->getbackground ();

	#else

	touch (ch);
	no_graphics_support ();
	throw ErrBlassicInternal;

	#endif
}

void blassic::graphics::cls (BlChannel n)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [n]->cls ();

	#else

	touch (n);
	no_graphics_support ();

	#endif
}

void blassic::graphics::definewindow (ScreenNum scnum, BlChannel n,
	int x1, int x2, int y1, int y2)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	touch (scnum);
	Screen * pcurrent= getcurrentscreen ();

	--x1; --x2; --y1; --y2;
	MapWindow::iterator it= mapwindow.find (n);
	if (it != mapwindow.end () )
	{
		BlTextWindow * pwin= it->second;
		if (pwin->get_screen () == pcurrent)
		{
			TRDEB ("Setting window");
			pwin->set (x1, x2, y1, y2);
			return;
		}
		else
		{
			delete pwin;
			mapwindow.erase (it);
		}
	}

	TRDEB ("New window");
	auto_ptr <BlTextWindow> pwin (new BlTextWindow
		(* pcurrent, x1, x2, y1, y2) );
	mapwindow.insert (make_pair (n, pwin.get () ) );
	pwin.release ();

	#else

	touch (scnum, n, x1, x2, y1, y2);

	#endif
}

void blassic::graphics::undefinewindow (BlChannel n)
{
	TRF;

	if (n == 0)
		return;

	#ifdef BLASSIC_HAS_GRAPHICS

	MapWindow::iterator it= mapwindow.find (n);
	if (it != mapwindow.end () )
	{
		delete it->second;
		mapwindow.erase (it);
	}

	#else

	ASSERT (false);
	no_graphics_support ();

	#endif
}

blassic::graphics::ScreenNum blassic::graphics::getscreennum (BlChannel ch)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	MapWindow::iterator it= mapwindow.find (ch);
	if (it == mapwindow.end () )
		throw ErrBlassicInternal;

	Screen * pscreen= it->second->get_screen ();
	return ScreenMap::find (pscreen);

	#else

	touch (ch);
	throw ErrBlassicInternal;

	#endif
}

size_t blassic::graphics::getlinewidth (BlChannel ch)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return mapwindow [ch]->getwidth ();

	#else

	touch (ch);
	no_graphics_support ();
	return 0; // Make the compiler happy

	#endif
}

void blassic::graphics::charout (BlChannel ch, char c)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->charout (c);

	#else

	touch (ch, c);

	#endif
}

void blassic::graphics::stringout (BlChannel ch, const std::string & str)
{
	TRFDEBS ("Channel: " << ch << " String: " << str);

	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->stringout (str);

	#else

	touch (ch, str);

	#endif
}

std::string blassic::graphics::copychr (BlChannel ch, BlChar from, BlChar to)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	BlTextWindow * pwin= mapwindow [ch];
	return pwin->copychr (from, to);

	#else

	touch (ch, from, to);
	ASSERT (false);
	throw ErrBlassicInternal;

	#endif
}

std::string blassic::graphics::screenchr (int x, int y)
{
	requiregraphics ();

	#if 0

	#ifdef BLASSIC_HAS_GRAPHICS

	return copychrat (x * charwidth, y * charheight, 0, 255);

	#else

	touch (x, y);
	ASSERT (false);
	throw ErrBlassicInternal;

	#endif

	#else

	#ifdef BLASSIC_HAS_GRAPHICS
	return getcurrentscreen ()->copychratpos (x, y, 0, 255);
	#else
	touch (x, y);
	return std::string ();
	#endif

	#endif
}

void blassic::graphics::gotoxy (BlChannel ch, int x, int y)
{
	TRFDEBS ("Channel: " << ch << ", x= " << x << ", y= " << y);

	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->gotoxy (x, y);

	#else

	touch (ch, x, y);

	#endif
}

void blassic::graphics::tab (BlChannel ch)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->tab ();

	#else

	touch (ch);

	#endif
}

void blassic::graphics::tab (BlChannel ch, size_t n)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->tab (n);

	#else

	touch (ch, n);

	#endif
}

void blassic::graphics::movecharforward (BlChannel ch, size_t n)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->movecharforward (n);

	#else

	touch (ch, n);

	#endif
}

void blassic::graphics::movecharback (BlChannel ch, size_t n)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->movecharback (n);

	#else

	touch (ch, n);

	#endif
}

void blassic::graphics::movecharup (BlChannel ch, size_t n)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->movecharup (n);

	#else

	touch (ch, n);

	#endif
}

void blassic::graphics::movechardown (BlChannel ch, size_t n)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->movechardown (n);

	#else

	touch (ch, n);

	#endif
}

void blassic::graphics::scroll (ScreenNum scnum,
	int x1, int y1, int x2, int y2, int xoff, int yoff,
	bool fill, int inknum)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	Screen * pscreen= ScreenMap::get (scnum);
	pscreen->scroll (x1, y1, x2, y2, xoff, yoff,
		fill, inknum);
	pscreen->flush ();

	#else

	touch (scnum, x1, y1, x2, y2);
	touch (xoff, yoff, fill, inknum);
	throw ErrFunctionCall;

	#endif
}

void blassic::graphics::scroll (BlChannel ch, int n)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->scroll (n);

	#else

	touch (ch, n);

	#endif
}

void blassic::graphics::symbolafter (int symbol)
{
	if (symbol < 0 || symbol > 256)
		throw ErrFunctionCall;

	#ifdef BLASSIC_HAS_GRAPHICS

	memcpy (charset::data, * charset::default_charset,
		sizeof (charset::chardataset) );
	symbol_after_is= symbol;

	#endif
}

void blassic::graphics::definesymbol (int symbol,
	const unsigned char (& byte) [8] )
{
	if (symbol < 0 || symbol > 255)
		throw ErrFunctionCall;

	#ifdef BLASSIC_HAS_GRAPHICS

	if (symbol < symbol_after_is)
		throw ErrImproperArgument;
	memcpy (charset::data + symbol, byte, sizeof (byte) );

	#else

	touch (byte);

	#endif
}

void blassic::graphics::synchronize (bool mode)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	Screen * pscreen= getcurrentscreen ();

	bool previous= pscreen->get_synchro ();
	TRDEBS ("Was " << previous);

	pscreen->set_synchro (mode);

	if (previous == true && mode == false && ingraphicsmode () )
		pscreen->update_window ();

	TRDEBS ("Set to " << mode);

	#else

	touch (mode);

	#endif
}

void blassic::graphics::synchronize ()
{
	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->update_window ();
	blassic::idle ();

	#endif
}

void blassic::graphics::synchronize_suspend ()
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->suspend_synchro ();

	#endif
}

void blassic::graphics::synchronize_restart ()
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->restart_synchro ();

	#endif
}

void blassic::graphics::synchronize_suspend (BlChannel ch)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->synchronize_suspend ();

	#else

	touch (ch);

	#endif
}

void blassic::graphics::synchronize_restart (BlChannel ch)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->synchronize_restart ();

	#else

	touch (ch);

	#endif
}

int blassic::graphics::xmouse ()
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return getcurrentscreen ()->get_xmouse ();

	#else

	return 0;

	#endif
}

int blassic::graphics::ymouse ()
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return getcurrentscreen ()->get_ymouse ();

	#else

	return 0;

	#endif
}

int blassic::graphics::xpos ()
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return getcurrentscreen ()->get_lastx ();

	#else

	return 0;

	#endif
}

int blassic::graphics::xpos (BlChannel ch)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return mapwindow [ch]->getxpos ();

	#else

	touch (ch);
	throw ErrFunctionCall;

	#endif
}

int blassic::graphics::ypos ()
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return getcurrentscreen ()->get_lasty ();

	#else

	return 0;

	#endif
}

int blassic::graphics::ypos (BlChannel ch)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return mapwindow [ch]->getypos ();

	#else

	touch (ch);
	throw ErrFunctionCall;

	#endif
}

void blassic::graphics::tag (BlChannel ch)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->tag ();

	#else

	touch (ch);

	#endif
}

void blassic::graphics::tagoff (BlChannel ch)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->tagoff ();

	#else

	touch (ch);

	#endif
}

bool blassic::graphics::istagactive (BlChannel ch)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return mapwindow [ch]->istagactive ();

	#else

	touch (ch);
	throw ErrFunctionCall;

	#endif
}

void blassic::graphics::showcursor (BlChannel ch)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->invertcursor ();

	#else

	touch (ch);

	#endif
}

void blassic::graphics::hidecursor (BlChannel ch)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->invertcursor ();

	#else

	touch (ch);

	#endif
}

void blassic::graphics::inverse (BlChannel ch, bool active)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->setinverse (active);

	#else

	touch (ch, active);

	#endif
}

bool blassic::graphics::getinverse (BlChannel ch)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return mapwindow [ch]->getinverse ();

	#else

	touch (ch);
	throw ErrFunctionCall;

	#endif
}

void blassic::graphics::bright (BlChannel ch, bool active)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	mapwindow [ch]->setbright (active);

	#else

	touch (ch, active);

	#endif
}

bool blassic::graphics::getbright (BlChannel ch)
{
	#ifdef BLASSIC_HAS_GRAPHICS

	return mapwindow [ch]->getbright ();

	#else

	touch (ch);
	throw ErrFunctionCall;

	#endif
}

void blassic::graphics::clean_input ()
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	QueueKey & queuekey= getcurrentscreen ()->getQueueKey ();

	blassic::idle ();
	queuekey.erase ();

	#endif
}

void blassic::graphics::ring ()
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->ring_my_bell ();

	#endif
}

void blassic::graphics::set_title (const std::string & title)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	getcurrentscreen ()->set_title (title);

	#else

	touch (title);

	#endif
}

void blassic::graphics::set_default_title (const std::string & title)
{
	TRF;

	#ifdef BLASSIC_HAS_GRAPHICS

	default_title= title;

	#else

	touch (title);

	#endif
}


//*********************************************************
//		Graphics get and put.
//*********************************************************


#ifdef BLASSIC_HAS_GRAPHICS


namespace {


#if 0
static int bitbltmodesbynumber []= { SRCCOPY, SRCINVERT,
	SRCAND, SRCPAINT, DSTINVERT };

int getbitbltmode (int mode)
{
	if (mode < 0 ||
		size_t (mode) >= dim_array (bitbltmodesbynumber) )
	{
		throw ErrFunctionCall;
	}
	return bitbltmodesbynumber [mode];
}
#endif


#if 0


class Image {
public:
	Image (int x1, int y1, int x2, int y2);
	~Image ();
	void put (int x, int y, int mode);
private:
	int width;
	int height;

	#ifdef BLASSIC_USE_X

	XImage * img;

	#elif defined BLASSIC_USE_WINDOWS_GRAPHICS

	HBITMAP img;
	HDC hdcImg;

	#endif
};

Image::Image (int x1, int y1, int x2, int y2)
{
	// Adjust coordinates to current rotate mode.
	do_rotate (x1, y1);
	do_rotate (x2, y2);
	if (x1 > x2) swap (x1, x2);
	if (y1 > y2) swap (y1, y2);
	if (x1 >= screenwidth || y1 >= screenheight)
		throw ErrFunctionCall;
	if (x2 >= screenwidth)
		x2= screenwidth - 1;
	if (y2 >= screenheight)
		y2= screenheight - 1;

	width= x2 - x1 + 1;
	height= y2 - y1 + 1;

	#ifdef BLASSIC_USE_X

	img= XGetImage (display, pixmap, x1, y1, width, height,
		AllPlanes, XYPixmap);
	if (img == NULL)
		throw ErrFunctionCall;

	#elif defined BLASSIC_USE_WINDOWS_GRAPHICS

	img= CreateCompatibleBitmap (hdcPixmap, width, height);
	if (img == NULL)
		throw ErrFunctionCall;
	hdcImg= CreateCompatibleDC (hdcPixmap);
	if (hdcImg == NULL)
	{
		DeleteObject (img);
		throw ErrFunctionCall;
	}
	HGDIOBJ hgdiobj= SelectObject (hdcImg, img);
	if (hgdiobj == NULL ||
		hgdiobj == reinterpret_cast <HGDIOBJ> (GDI_ERROR) )
	{
		DeleteDC (hdcImg);
		DeleteObject (img);
		throw ErrFunctionCall;
	}

	if (BitBlt (hdcImg, 0, 0, width, height, hdcPixmap, x1, y1, SRCCOPY)
		== 0)
	{
		DeleteDC (hdcImg);
		DeleteObject (img);
		throw ErrFunctionCall;
	}

	#endif
}

Image::~Image ()
{
	#ifdef BLASSIC_USE_X

	XDestroyImage (img);

	#elif defined BLASSIC_USE_WINDOWS_GRAPHICS

	DeleteDC (hdcImg);
	DeleteObject (img);

	#endif
}

void Image::put (int x, int y, int mode)
{
	TRACEFUNC (tr, "Image::put");

	TRMESSAGE (tr, "at " + to_string (x) + ',' + to_string (y) );

	// Adjust coordinates to current rotate mode.
	do_rotate (x, y);
	switch (rotate)
	{
	case RotateNone:
		// Nothing to do.
		break;
	case Rotate90:
		y-= width - 1;
		break;
	}
	if (x >= screenwidth || y >= screenheight)
		throw ErrFunctionCall;
	int maxwidth= screenwidth - x + 1;
	int maxheight= screenheight - y + 1;
	int w= std::min (width, maxwidth);
	int h= std::min (height, maxheight);

	TRMESSAGE (tr, "At " + to_string (x) + ',' + to_string (y) );

	#ifdef BLASSIC_USE_X

	int modeused= getdrawmode (mode);

	XSetFunction (display, gcp, modeused);
	XPutImage (display, pixmap, gcp, img,
		0, 0, x, y, w, h);
	XSetFunction (display, gcp, drawmode);

	#elif defined BLASSIC_USE_WINDOWS_GRAPHICS

	int modeused= getbitbltmode (mode);

	if (BitBlt (hdcPixmap, x, y, w, h, hdcImg, 0, 0, modeused) == 0)
		throw ErrFunctionCall;

	#endif

	if (! fSynchro)
		update_window (x, y, width, height);
}


typedef map <std::string, Image *> imagemap_t;
imagemap_t imagemap;


#endif


} // namespace

#endif
// BLASSIC_HAS_GRAPHICS


void blassic::graphics::get_image (int x1, int y1, int x2, int y2,
	const std::string & name)
{
	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	imagemap_t::iterator it= imagemap.find (name);
	if (it != imagemap.end () )
	{
		delete it->second;
		imagemap.erase (it);
	}

	#if 0
	auto_ptr <Image> pimg (new Image (x1, y1, x2, y2) );
	#endif

	auto_ptr <Image>
		pimg (getcurrentscreen ()->getImage (x1, y1, x2, y2) );

	imagemap [name]= pimg.get ();
	pimg.release ();

	#else
	// No BLASSIC_HAS_GRAPHICS

	touch (x1, y1, x2, y2, name);

	#endif
}

void blassic::graphics::put_image (int x, int y,
	const std::string & name, int mode)
{
	requiregraphics ();

	#ifdef BLASSIC_HAS_GRAPHICS

	imagemap_t::iterator it= imagemap.find (name);
	if (it != imagemap.end () )
	{
		#if 0
		it->second->put (x, y, mode);
		#endif

		getcurrentscreen ()->putImage (* it->second, x, y, mode);
	}
	else
		throw ErrFunctionCall;

	#else

	touch (x, y, name, mode);

	#endif
	// BLASSIC_HAS_GRAPHICS
}

void blassic::graphics::clear_images ()
{
	#ifdef BLASSIC_HAS_GRAPHICS

	for (imagemap_t::iterator it= imagemap.begin ();
		it != imagemap.end ();
		++it)
	{
		delete it->second;
	}
	imagemap.clear ();

	#endif
	// BLASSIC_HAS_GRAPHICS
}

void blassic::idle ()
{
	#ifdef BLASSIC_HAS_GRAPHICS

	using namespace blassic::graphics_impl;

	//if (! window_created)
	//	return;

	//getcurrentscreen ()->check_events ();

	#ifndef BLASSIC_HAS_THREADS
	SystemMap::check_events ();
	#endif

	#endif
	// BLASSIC_HAS_GRAPHICS
}


// Fin de graphics.cpp
